_filenames = {
    "data.csv": {
        "url": "http://rest.s3for.me/limix/example/data.csv",
        "sha256": "9b09e5d569b256aabc42d3e04139f7f55789283ee642023adee455b6e5ef7d14",
    },
    "data.h5": {
        "url": "http://rest.s3for.me/limix/example/data.h5",
        "sha256": "30bc0374211f4ad7f763f631fc7cf8d1fdee78d5a0064facc8fcf09ab9ace64c",
    },
    "example.gen": {
        "url": "http://rest.s3for.me/limix/example/example.gen",
        "sha256": "a21903b261254981edf552798bda8f911423bf92083759b93f3e78295cc7e382",
    },
    "example.sample": {
        "url": "http://rest.s3for.me/limix/example/example.sample",
        "sha256": "f10f3304104fbecfd15703da06ef3e9d922a5dc069644bacb3f952d568d4c9db",
    },
    "expr.csv": {
        "url": "http://rest.s3for.me/limix/example/expr.csv",
        "sha256": "48ab488b1a1a7651d1152dae1e13605df9a342667ccba48328ae4b86b3ac359a",
    },
    "pheno.csv": {
        "url": "http://rest.s3for.me/limix/example/pheno.csv",
        "sha256": "9b09e5d569b256aabc42d3e04139f7f55789283ee642023adee455b6e5ef7d14",
    },
    "xarr.hdf5": {
        "url": "http://rest.s3for.me/limix/example/xarr.hdf5",
        "sha256": "fde8e4830d04d47f4463ba17704679b3259dc4907f56cc2d69a930a584ea6a00",
    },
    "phenotype.gemma": {
        "url": "http://rest.s3for.me/limix/example/phenotype.gemma",
        "sha256": "90089e70364dc5e2b6d71ea00e3b5e60203d9940ca909cf2ce688aca5178044b",
    },
    "data.bed": {
        "url": "http://rest.s3for.me/limix/example/data.bed",
        "sha256": "7bf9162d486116db12686e14ef7bf2d53c46bb788041555041891cebabf1ba09",
    },
    "data.bim": {
        "url": "http://rest.s3for.me/limix/example/data.bim",
        "sha256": "f6b70bebe81aad8941903cd2a53bcf30507efdd7c52d6692e8d62cb90a7ddb1c",
    },
    "data.fam": {
        "url": "http://rest.s3for.me/limix/example/data.fam",
        "sha256": "701a94fc09860bc341b42497ac9cae893e771798270d69d3fb566cfa09c03c8f",
    },
    "data.h5.bz2": {
        "url": "http://rest.s3for.me/limix/example/data.h5.bz2",
        "sha256": "2df755158b403416e1b0270815b2fc18f7d23ec5590c9cabacd2e1be98510c59",
    },
    "chrom22_subsample20_maf0.10.bed": {
        "url": "http://rest.s3for.me/limix/example/chrom22_subsample20_maf0.10.bed",
        "sha256": "cf18099968dbb9a514d2e9b86f02f2fab2c549e7859d7560ae5bc79a841dd269",
    },
    "chrom22_subsample20_maf0.10.fam": {
        "url": "http://rest.s3for.me/limix/example/chrom22_subsample20_maf0.10.fam",
        "sha256": "76bd241625bf2b6aa5fe763d4d037609fbd8c70eadf261faeab99cdd5fe7443e",
    },
    "chrom22_subsample20_maf0.10.bim": {
        "url": "http://rest.s3for.me/limix/example/chrom22_subsample20_maf0.10.bim",
        "sha256": "24824433bef33e17a9f9713868f492c9957efbc585e0010dfcc8536fdf7996d4",
    },
    "expr_nan.csv": {
        "url": "http://rest.s3for.me/limix/example/expr_nan.csv",
        "sha256": "6ec7d2a21dbb2b2e65b0a28cf7f7bdbf6cbb974f9cfd1d653050e6b1aa84dfbb",
    },
}


def _download(url, filepath, hash):
    import os
    from urllib.request import urlretrieve
    import limix

    local_filename = urlretrieve(url)[0]
    os.rename(local_filename, filepath)
    if limix.sh.filehash(filepath) != hash:
        raise RuntimeError("Hash does not match for file {}.".format(filepath))


class file_example(object):
    """
    Put file examples into a temporary folder.
    """

    def __init__(self, filenames):
        import tempfile
        from limix.sh._dir import makedirs
        import os
        from os.path import join
        from limix.sh._user_dir import user_cache_dir
        import limix

        self._unlist = False
        if not isinstance(filenames, (tuple, list)):
            filenames = [filenames]
            self._unlist = True

        self._orig_folder = join(user_cache_dir(), "examples")
        makedirs(self._orig_folder)

        for fn in filenames:
            if fn not in _filenames:
                raise ValueError(f"Unrecognized file name {fn}.")
            info = _filenames[fn]
            fp = join(self._orig_folder, fn)
            if os.path.exists(fp):
                if limix.sh.filehash(fp) != info["sha256"]:
                    os.remove(fp)
                    _download(info["url"], fp, info["sha256"])
            else:
                _download(info["url"], fp, info["sha256"])

        self._dirpath = tempfile.mkdtemp()
        self._filenames = filenames

    def __enter__(self):
        import shutil
        from os.path import join

        filepaths = []
        for fn in self._filenames:
            fp = join(self._dirpath, fn)
            shutil.copy(join(self._orig_folder, fn), fp)
            filepaths.append(fp)

        if self._unlist:
            return filepaths[0]
        return filepaths

    def __exit__(self, *_):
        import shutil

        try:
            shutil.rmtree(self._dirpath)
        except PermissionError:
            pass



from ._mt_result import MTScanResult
from ._mt_simple import MTSimpleModelResult


class MTScanResultFactory:
    def __init__(
        self,
        lik,
        traits,
        covariates,
        candidates,
        envs0,
        envs1,
        lml,
        beta,
        beta_se,
        C0,
        C1,
    ):
        from numpy import asarray, atleast_1d

        self._h0 = MTSimpleModelResult(
            lik, traits, covariates, lml, beta, beta_se, C0, C1
        )
        self._tests = []
        self._traits = asarray(atleast_1d(traits), str)
        self._covariates = asarray(atleast_1d(covariates), str)
        self._candidates = asarray(atleast_1d(candidates), str)
        self._envs0 = asarray(atleast_1d(envs0), str)
        self._envs1 = asarray(atleast_1d(envs1), str)

    def add_test(self, cand_idx, h1, h2):
        from numpy import atleast_1d, atleast_2d, asarray

        if not isinstance(cand_idx, slice):
            cand_idx = asarray(atleast_1d(cand_idx).ravel(), int)

        def _2d_shape(x):
            x = asarray(x, float)
            x = atleast_2d(x.T).T
            return x

        def _normalize(h):
            return {
                "lml": float(h["lml"]),
                "covariate_effsizes": _2d_shape(h["covariate_effsizes"]),
                "candidate_effsizes": _2d_shape(h["candidate_effsizes"]),
                "covariate_effsizes_se": _2d_shape(h["covariate_effsizes_se"]),
                "candidate_effsizes_se": _2d_shape(h["candidate_effsizes_se"]),
                "scale": float(h["scale"]),
            }

        r = {"idx": cand_idx, "h2": _normalize(h2)}
        if h1 is not None:
            r["h1"] = _normalize(h1)

        self._tests.append(r)

    def create(self):
        return MTScanResult(
            self._tests,
            self._traits,
            self._covariates,
            self._candidates,
            self._h0,
            self._envs0,
            self._envs1,
        )



import warnings

from limix._bits import dask, numpy, pandas, xarray


def mean_standardize(X, axis=-1, inplace=False):
    """
    Zero-mean and one-deviation normalisation.

    Normalise in such a way that the mean and variance are equal to zero and
    one. This transformation is taken over the flattened array by default, otherwise
    over the specified axis. Missing values represented by ``NaN`` are ignored.

    Parameters
    ----------
    X : array_like
        Array of values.
    axis : int, optional
        Axis value. Defaults to ``1``.
    inplace : bool, optional
        Defaults to ``False``.

    Returns
    -------
    X : ndarray
        Normalized array.

    Example
    -------
    .. doctest::

        >>> import limix
        >>> from numpy import arange
        >>>
        >>> X = arange(15).reshape((5, 3)).astype(float)
        >>> print(X)
        [[ 0.  1.  2.]
         [ 3.  4.  5.]
         [ 6.  7.  8.]
         [ 9. 10. 11.]
         [12. 13. 14.]]
        >>> X = arange(6).reshape((2, 3)).astype(float)
        >>> X = limix.qc.mean_standardize(X, axis=0)
        >>> print(X) # doctest: +FLOAT_CMP
        [[-1.22474487  0.          1.22474487]
         [-1.22474487  0.          1.22474487]]
    """

    from numpy import issubdtype, integer, asarray

    if hasattr(X, "dtype") and issubdtype(X.dtype, integer):
        raise ValueError("Integer type is not supported.")

    if isinstance(X, (tuple, list)):
        if inplace:
            raise ValueError("Can't use `inplace=True` for {}.".format(type(X)))
        X = asarray(X, float)

    if numpy.is_array(X):
        X = _mean_standardize_numpy(X, axis, inplace)
    elif pandas.is_series(X):
        X = _mean_standardize_pandas_series(X, axis, inplace)
    elif pandas.is_dataframe(X):
        X = _mean_standardize_pandas_dataframe(X, axis, inplace)
    elif dask.is_array(X):
        X = _mean_standardize_dask_array(X, axis, inplace)
    elif dask.is_series(X):
        raise NotImplementedError()
    elif dask.is_dataframe(X):
        X = _mean_standardize_dask_dataframe(X, axis, inplace)
    elif xarray.is_dataarray(X):
        X = _mean_standardize_xarray_dataarray(X, axis, inplace)
    else:
        raise NotImplementedError()

    return X


def _mean_standardize_numpy(X, axis, inplace):
    from numpy import nanmean, nanstd
    from numpy_sugar import epsilon
    from numpy import inf, clip

    orig_shape = X.shape
    if X.ndim == 1:
        X = X.reshape(orig_shape + (1,))

    if not inplace:
        X = X.copy()

    X = X.swapaxes(1, axis)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        X -= nanmean(X, axis=0)
        X /= clip(nanstd(X, axis=0), epsilon.tiny, inf)

    X = X.swapaxes(1, axis)

    return X.reshape(orig_shape)


def _mean_standardize_pandas_series(X, axis, inplace):
    if not inplace:
        X = X.copy()

    a = X.to_numpy()
    _mean_standardize_numpy(a, axis, True)
    X[:] = a

    return X


def _mean_standardize_pandas_dataframe(x, axis, inplace):
    if not inplace:
        x = x.copy()

    a = x.to_numpy()
    _mean_standardize_numpy(a, axis, True)
    x[:] = a

    return x


def _mean_standardize_dask_array(x, axis, inplace):
    import dask.array as da
    from numpy_sugar import epsilon
    from numpy import nanmean, clip, nanstd, inf

    if inplace:
        raise NotImplementedError()

    x = x.swapaxes(1, axis)

    x = dask.array_shape_reveal(x)
    shape = da.compute(*x.shape)

    def func(a):
        a -= nanmean(a, axis=0)
        a /= clip(nanstd(a, axis=0), epsilon.tiny, inf)
        return a

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        x = da.apply_along_axis(_dask_apply, 0, x, func, shape[0])

    return x.swapaxes(1, axis)


def _mean_standardize_dask_dataframe(x, axis, inplace):
    if inplace:
        raise NotImplementedError()

    d = x.to_dask_array(lengths=True)
    orig_chunks = d.chunks
    d = _mean_standardize_dask_array(d, axis, False).rechunk(orig_chunks)
    return d.to_dask_dataframe(columns=x.columns, index=x.index)


def _mean_standardize_xarray_dataarray(X, axis, inplace):
    if not inplace:
        X = X.copy(deep=True)

    data = X.data

    if dask.is_array(data):
        data = _mean_standardize_dask_array(data, axis, inplace)
    else:
        data = _mean_standardize_numpy(data, axis, inplace)

    X.data = data
    return X


def _dask_apply(x, func1d, length):
    from numpy import resize

    x = func1d(x)
    return resize(x, length)



class Chi2Mixture(object):
    """
    Mixture of 𝜒² distributions.

    Class for evaluation of P values for a test statistic that follows a
    two-component mixture of chi2

    .. math::

        p(x) = (1-𝑝)𝜒²(0) + 𝑝𝑎𝜒²(𝑑).

    Here 𝑝 is the probability being in the first component and 𝑎 and 𝑑 are the scale
    parameter and the number of degrees of freedom of the second
    component.

    .. warning::
        This class is not production-ready. Keep in mind that this interface is likely
        to change.

    Parameters
    ----------
    scale_min : float
        Minimum value used for fitting the scale parameter.
    scale_max : float
        Maximum value used for fitting the scale parameter.
    dofmin : float
        Minimum value used for fitting the dof parameter.
    dofmax : float
        Maximum value used for fitting the dof parameter.
    qmax : float
        Only the top qmax quantile is used for the fit.
    n_interval : int
        Number of intervals when performing gridsearch.
    tol : float
        Tolerance of being zero.

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>> import scipy as sp
        >>> from limix.stats import Chi2Mixture
        >>>
        >>> scale = 0.3
        >>> dof = 2
        >>> mixture = 0.2
        >>> n = 100
        >>>
        >>> random = RandomState(1)
        >>> x =  random.chisquare(dof, n)
        >>> n0 = int( (1-mixture) * n)
        >>> idxs = random.choice(n, n0, replace=False)
        >>> x[idxs] = 0
        >>>
        >>> chi2mix = Chi2Mixture(scale_min=0.1, scale_max=5.0,
        ...                       dof_min=0.1, dof_max=5.0,
        ...                       qmax=0.1, tol=4e-3)
        >>> chi2mix.estimate_chi2mixture(x)
        >>> pv = chi2mix.sf(x)
        >>> print(pv[:4]) # doctest: +FLOAT_CMP
        [0.2 0.2 0.2 0.2]
        >>>
        >>> print('%.2f' % chi2mix.scale)
        1.98
        >>> print('%.2f' % chi2mix.dof)
        0.89
        >>> print('%.2f' % chi2mix.mixture)
        0.20
    """

    def __init__(
        self,
        scale_min=0.1,
        scale_max=5.0,
        dof_min=0.1,
        dof_max=5.0,
        n_intervals=100,
        qmax=0.1,
        tol=0,
        lrt=None,
    ):

        self.scale_min = scale_min
        self.scale_max = scale_max
        self.dof_min = dof_min
        self.dof_max = dof_max
        self.qmax = qmax
        self.n_intervals = n_intervals
        self.tol = tol

    def estimate_chi2mixture(self, lrt):
        r"""Estimates the parameters of a chi2 mixture.

        Estimates the parameters of a chi2 mixture by fitting the empirical
        distribution of null test statistic.

        Parameters
        ----------
        lrt : array_like
            Null test statistcs.
        """
        import scipy as sp
        import scipy.stats as st

        # step 1: estimate the probability of being in component one
        self.mixture = 1 - (lrt <= self.tol).mean()
        n_false = sp.sum(lrt > self.tol)

        # step 2: only use the largest qmax fraction of test statistics to
        #         estimate the remaining parameters
        n_fitting = int(sp.ceil(self.qmax * n_false))
        lrt_sorted = -sp.sort(-lrt)[:n_fitting]
        q = sp.linspace(0, 1, n_false)[1 : n_fitting + 1]
        log_q = sp.log10(q)

        # step 3: fitting scale and dof by minimizing the squared error
        #          of the log10 p-values with their theorietical values
        #          [uniform distribution]
        MSE_opt = sp.inf
        MSE = sp.zeros((self.n_intervals, self.n_intervals))

        for i, scale in enumerate(
            sp.linspace(self.scale_min, self.scale_max, self.n_intervals)
        ):
            for j, dof in enumerate(
                sp.linspace(self.dof_min, self.dof_max, self.n_intervals)
            ):
                p = st.chi2.sf(lrt_sorted / scale, dof)
                log_p = sp.log10(p)
                MSE[i, j] = sp.mean((log_q - log_p) ** 2)
                if MSE[i, j] < MSE_opt:
                    MSE_opt = MSE[i, j]
                    self.scale = scale
                    self.dof = dof

    def sf(self, lrt):
        r"""Computes the p-values from test statistics lrt.

        Parameters
        ----------
        lrt : array_like
            Test statistics.

        Returns
        -------
        array_like
            P-values.
        """
        from numpy import asarray
        import scipy as sp
        import scipy.stats as st

        lrt = asarray(lrt, float)

        _lrt = sp.copy(lrt)
        _lrt[lrt < self.tol] = 0
        pv = self.mixture * st.chi2.sf(_lrt / self.scale, self.dof)
        return pv



from os.path import exists

recognized_file_types = [
    "image",
    "hdf5",
    "csv",
    "npy",
    "grm.raw",
    "bed",
    "bgen",
    "bimbam-pheno",
]


def infer_filetype(filepath):
    imexts = [".png", ".bmp", ".jpg", "jpeg"]
    if filepath.endswith(".hdf5") or filepath.endswith(".h5"):
        return "hdf5"
    if filepath.endswith(".csv"):
        return "csv"
    if filepath.endswith(".npy"):
        return "npy"
    if filepath.endswith(".grm.raw"):
        return "grm.raw"
    if _is_bed(filepath):
        return "bed"
    if any([filepath.endswith(ext) for ext in imexts]):
        return "image"
    if filepath.endswith(".txt"):
        return "csv"
    if filepath.endswith(".bgen"):
        return "bgen"
    if filepath.endswith(".gemma"):
        return "bimbam-pheno"
    return "unknown"


def _is_bed(filepath):
    files = [filepath + ext for ext in [".bed", ".bim", ".fam"]]
    ok = [exists(f) for f in files]

    if sum(ok) > 0 and sum(ok) < 3:
        mfiles = ", ".join([files[i] for i in range(3) if not ok[i]])
        print("The following file(s) are missing:", mfiles)
        return False

    return all(ok)



import dask.array as da
import dask.dataframe as dd
import pandas as pd
import xarray as xr
from numpy.testing import assert_, assert_allclose, assert_array_equal


def assert_ndarray_1d(func, data1d):
    _assert_mat_proc(func, data1d, lambda X, *_: X.copy())
    _assert_mat_proc_inplace(func, data1d, lambda X, *_: X.copy())


def assert_ndarray_2d(func, data2d):
    _assert_mat_proc(func, data2d, lambda X, *_: X.copy())
    _assert_mat_proc_inplace(func, data2d, lambda X, *_: X.copy())


def assert_pandas_series(func, data1d):
    def get(X, samples, *_):
        return pd.Series(X, index=samples)

    _assert_mat_proc(func, data1d, get)
    _assert_mat_proc_inplace(func, data1d, get)


def assert_pandas_dataframe(func, data2d):
    def get(X, samples, candidates):
        return pd.DataFrame(X, index=samples, columns=candidates)

    _assert_mat_proc(func, data2d, get)
    _assert_mat_proc_inplace(func, data2d, get)


def assert_dask_array(func, data2d, test_inplace=True):
    def get(X, *_):
        return da.from_array(X.copy(), chunks=2)

    _assert_mat_proc(func, data2d, get)
    if test_inplace:
        _assert_mat_proc_inplace(func, data2d, get)


def assert_dask_dataframe(func, data2d):
    def get(X, samples, candidates):
        df = pd.DataFrame(X.copy(), index=samples, columns=candidates)
        df = dd.from_pandas(df, npartitions=2)
        return df

    _assert_mat_proc(func, data2d, get)


def assert_xarray_dataarray(func, data2d, test_inplace=True):
    def get(X, samples, candidates):
        X = xr.DataArray(
            X.copy(),
            coords={"sample": samples, "candidate": candidates},
            dims=["sample", "candidate"],
        )
        return X

    _assert_mat_proc(func, data2d, get)
    if test_inplace:
        _assert_mat_proc_inplace(func, data2d, get)


def _assert_mat_proc(func, data, astype):
    def get():
        return astype(data["X"], data["samples"], data["candidates"]).copy()

    X = get()
    assert_allclose(func(X), data["R"])
    assert_allclose(X, get())
    assert_(isinstance(func(X), type(get())))
    _assert_attr_values(X, get())

    X = get()
    assert_allclose(func(X, axis=0), data["Rt"])
    assert_allclose(X, get())
    assert_(isinstance(func(X, axis=0), type(get())))
    _assert_attr_values(X, get())


def _assert_mat_proc_inplace(func, data, astype):
    def get():
        return astype(data["X"], data["samples"], data["candidates"]).copy()

    X = get().copy()
    assert_allclose(func(X, inplace=True), data["R"])
    assert_allclose(X, data["R"])
    assert_(isinstance(func(X, inplace=True), type(get())))
    _assert_attr_values(X, get())

    X = get().copy()
    assert_allclose(func(X, axis=0, inplace=True), data["Rt"])
    assert_allclose(X, data["Rt"])
    assert_(isinstance(func(X, axis=0, inplace=True), type(get())))
    _assert_attr_values(X, get())


def _assert_attr_values(X, R):
    for attr in ["index", "columns"]:
        if hasattr(X, attr):
            assert_array_equal(getattr(X, attr), getattr(R, attr))



import pytest
from numpy import asarray, nan
from numpy.random import RandomState

from limix.qc import quantile_gaussianize as qg

from .util import (
    assert_dask_array,
    assert_dask_dataframe,
    assert_ndarray_1d,
    assert_ndarray_2d,
    assert_pandas_dataframe,
    assert_pandas_series,
    assert_xarray_dataarray,
)


@pytest.fixture
def data2d():
    random = RandomState(0)
    X = random.randn(5, 2)
    X[0, 0] = nan

    R = asarray(
        [
            [nan, 0.0],
            [0.253_347_103, 0.967_421_566],
            [0.841_621_234, -0.967_421_566],
            [-0.253_347_103, -0.430_727_299],
            [-0.841_621_234, 0.430_727_299],
        ]
    )

    Rt = asarray(
        [
            [nan, 0.0],
            [-0.430_727_299, 0.430_727_299],
            [0.430_727_299, -0.430_727_299],
            [0.430_727_299, -0.430_727_299],
            [-0.430_727_299, 0.430_727_299],
        ]
    )

    samples = [f"sample{i}" for i in range(X.shape[0])]
    candidates = [f"snp{i}" for i in range(X.shape[1])]

    return {"X": X, "R": R, "Rt": Rt, "samples": samples, "candidates": candidates}


@pytest.fixture
def data1d():
    X = asarray([nan, 0.400_157_208_367_223_3, 0.978_737_984_105_739_2])
    R = asarray([nan, -0.430_727_299, 0.430_727_299])
    Rt = [nan, 0, 0]
    samples = [f"sample{i}" for i in range(X.shape[0])]
    return {"X": X, "R": R, "Rt": Rt, "samples": samples, "candidates": None}


def test_impute_ndarray_1d(data1d):
    assert_ndarray_1d(qg, data1d)


def test_impute_ndarray_2d(data2d):
    assert_ndarray_2d(qg, data2d)


def test_impute_pandas_series(data1d):
    assert_pandas_series(qg, data1d)


def test_impute_pandas_dataframe(data2d):
    assert_pandas_dataframe(qg, data2d)


def test_impute_dask_array(data2d):
    assert_dask_array(qg, data2d, test_inplace=False)


def test_impute_dask_dataframe(data2d):
    assert_dask_dataframe(qg, data2d)


def test_impute_xarray_dataarray(data2d):
    assert_xarray_dataarray(qg, data2d)



def assert_finite(Y, M, K):
    from numpy_sugar import is_all_finite

    if not is_all_finite(Y):
        raise ValueError("Outcome must have finite values only.")

    if not is_all_finite(M):
        raise ValueError("Covariates must have finite values only.")

    if K is not None:
        if not is_all_finite(K):
            raise ValueError("Covariate matrix must have finite values only.")



from dask.array import from_array
from numpy.random import RandomState
from numpy.testing import assert_equal

from limix.qc import indep_pairwise


def test_indep_pairwise():
    random = RandomState(0)

    X = random.randn(3, 10)

    head = [True, True, False, False, False]
    tail = [False, True, False, False]

    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[:5], head)
    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[-4:], tail)

    X = from_array(X, chunks=(2, 10))
    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[:5], head)
    assert_equal(indep_pairwise(X, 4, 2, 0.5, verbose=False)[-4:], tail)



def assert_target(target):
    from ._conf import CONF

    if target not in CONF["targets"]:
        raise ValueError(f"Unknown target `{target}`.")


def assert_filetype(filetype):
    from ._conf import CONF

    if filetype not in CONF["filetypes"]:
        raise ValueError(f"Unknown filetype `{filetype}`.")


def assert_likelihood(likname):
    from ._conf import CONF

    if likname not in CONF["likelihoods"]:
        msg = "Unrecognized likelihood name: {}.\n".format(likname)
        msg += "Valid names are: {}.".format(list(CONF["likelihoods"]))
        raise ValueError(msg)



def unique_variants(X):
    """
    Filters out variants with the same genetic profile.

    Parameters
    ----------
    X : array_like
        Samples-by-variants matrix of genotype values.

    Returns
    -------
    genotype : ndarray
        Genotype array with unique variants.

    Example
    -------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from numpy import kron, ones, sort
        >>> from limix.qc import unique_variants
        >>> random = RandomState(1)
        >>>
        >>> N = 4
        >>> X = kron(random.randn(N, 3) < 0, ones((1, 2)))
        >>>
        >>> print(X)
        [[0. 0. 1. 1. 1. 1.]
         [1. 1. 0. 0. 1. 1.]
         [0. 0. 1. 1. 0. 0.]
         [1. 1. 0. 0. 1. 1.]]
        >>>
        >>> print(unique_variants(X))
        [[0. 1. 1.]
         [1. 1. 0.]
         [0. 0. 1.]
         [1. 1. 0.]]
    """
    from limix._bits import dask
    import numpy as np

    if dask.is_array(X):
        import dask.array as da

        dot = da.dot
        unique = da.unique
    else:

        dot = np.dot
        unique = np.unique

    u = np.random.rand(X.shape[0])
    i = unique(dot(u, X), return_index=True)[1]
    return X[:, i]



def count_missingness(X):
    """
    Count the number of missing values per column.

    Parameters
    ----------
    X : array_like
        Matrix.

    Returns
    -------
    count : ndarray
        Number of missing values per column.
    """
    import dask.array as da
    from numpy import isnan

    if isinstance(X, da.Array):
        return da.isnan(X).sum(axis=0).compute()

    return isnan(X).sum(axis=0)



from limix._cache import cache


class STSimpleModelResult:
    def __init__(self, lik, trait, covariates, lml, beta, beta_se, v0, v1):
        from numpy import asarray, atleast_1d

        self._lik = lik
        self._trait = str(trait)
        self._covariates = asarray(atleast_1d(covariates), str)
        self._lml = float(lml)
        self._beta = asarray(beta, float).ravel()
        self._beta_se = asarray(beta_se, float).ravel()
        self._v0 = float(v0)
        self._v1 = float(v1)

    @property
    def trait(self):
        return self._trait

    @property
    def likelihood(self):
        return self._lik

    @property
    def lml(self):
        return self._lml

    @property
    def effsizes(self):
        return self._dataframes["effsizes"]

    @property
    def variances(self):
        return self._dataframes["variances"]

    @property
    @cache
    def _dataframes(self):
        from pandas import DataFrame

        effsizes = []
        for j, c in enumerate(self._covariates):
            effsizes.append([self._trait, c, self._beta[j], self._beta_se[j]])

        columns = ["trait", "covariate", "effsize", "effsize_se"]
        df0 = DataFrame(effsizes, columns=columns)

        variances = [[self._trait, self._v0, self._v1]]
        columns = ["trait", "fore_covariance", "back_covariance"]
        df1 = DataFrame(variances, columns=columns)

        return {"effsizes": df0, "variances": df1}



from __future__ import division


def linear_kinship(G, out=None, verbose=True):
    """
    Estimate Kinship matrix via linear kernel.

    Let 𝑑 be the number of columns of ``G``. The resulting matrix is given by:

    .. math::

        𝙺 = 𝚇𝚇ᵀ/𝑑

    where

    .. math::

        𝚇ᵢⱼ = (𝙶ᵢⱼ - 𝑚ⱼ) / 𝑠ⱼ

    is the matrix ``G`` column-wise normalized by means 𝑚ⱼ and standard deviations 𝑠ⱼ.
    NaNs are ignored so as to produce matrix ``K`` having only real numbers.

    This functions is specially designed to also handle large matrices ``G`` that would
    otherwise require a large amount of memory if it were to be loaded in memory first.
    For those cases, libraries like Dask come in handy.

    Parameters
    ----------
    G : array_like
        Samples-by-variants matrix.
    out : ndarray
        A location into which the result is stored.
    verbose : bool, optional
        ``True`` for showing progress; ``False`` otherwise. Defauts to ``True``.

    Examples
    --------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.stats import linear_kinship
        >>>
        >>> random = RandomState(1)
        >>> X = random.randn(4, 100)
        >>> K = linear_kinship(X, verbose=False)
        >>> print(K) # doctest: +FLOAT_CMP
        [[ 0.91314823 -0.19283362 -0.34133897 -0.37897564]
         [-0.19283362  0.89885153 -0.2356003  -0.47041761]
         [-0.34133897 -0.2356003   0.95777313 -0.38083386]
         [-0.37897564 -0.47041761 -0.38083386  1.23022711]]
    """
    from numpy import sqrt, zeros, asfortranarray, where, asarray, nanmean, std, isnan
    from scipy.linalg.blas import get_blas_funcs
    from tqdm import tqdm

    (n, p) = G.shape
    if out is None:
        out = zeros((n, n), order="F")
    else:
        out = asfortranarray(out)

    chunks = _get_chunks(G)
    gemm = get_blas_funcs("gemm", [out])

    start = 0
    for chunk in tqdm(chunks, desc="Kinship", disable=not verbose):
        end = start + chunk

        g = asarray(G[:, start:end])
        m = nanmean(g, 0)
        g = where(isnan(g), m, g)
        g = g - m
        g /= std(g, 0)
        g /= sqrt(p)

        gemm(1.0, g, g, 1.0, out, 0, 1, 1)

        start = end

    return out


def _get_chunks(G):
    from numpy import isfinite

    if hasattr(G, "chunks") and G.chunks is not None:
        if len(G.chunks) > 1 and all(isfinite(G.chunks[0])):
            return G.chunks[1]

    siz = G.shape[1] // 100
    sizl = G.shape[1] - siz * 100
    chunks = [siz] * 100
    if sizl > 0:
        chunks += [sizl]
    return tuple(chunks)



from ._st_result import STScanResult
from ._st_simple import STSimpleModelResult


class STScanResultFactory:
    def __init__(self, lik, trait, covariates, candidates, lml, beta, beta_se, v0, v1):
        from numpy import asarray, atleast_1d

        self._h0 = STSimpleModelResult(
            lik, trait, covariates, lml, beta, beta_se, v0, v1
        )
        self._tests = []
        self._trait = str(trait)
        self._covariates = asarray(atleast_1d(covariates), str)
        self._candidates = asarray(atleast_1d(candidates), str)

    def add_test(self, cand_idx, h2):
        from numpy import atleast_1d, asarray

        if not isinstance(cand_idx, slice):
            cand_idx = asarray(atleast_1d(cand_idx).ravel(), int)

        def _1d_shape(x):
            x = asarray(x, float)
            x = atleast_1d(x.T).T
            return x

        def _normalize(h):
            return {
                "lml": float(h["lml"]),
                "covariate_effsizes": _1d_shape(h["covariate_effsizes"]),
                "candidate_effsizes": _1d_shape(h["candidate_effsizes"]),
                "covariate_effsizes_se": _1d_shape(h["covariate_effsizes_se"]),
                "candidate_effsizes_se": _1d_shape(h["candidate_effsizes_se"]),
                "scale": float(h["scale"]),
            }

        self._tests.append({"idx": cand_idx, "h2": _normalize(h2)})

    def create(self):
        return STScanResult(
            self._tests, self._trait, self._covariates, self._candidates, self._h0
        )



from .._display import session_line


def download(url, dest=None, verbose=True):
    """
    Download file.

    Parameters
    ----------
    url : str
        Url to the file.
    dest : str, optional
        File destination. The current working directory is used if ``None`` is passed.
        Defaults to ``None``.
    verbose : bool, optional
        ``True`` for displaying progress. Defaults to ``True``.

    Returns
    -------
    filepath : str
        File path to the downloaded file.
    """
    import os
    from urllib.request import urlretrieve

    if dest is None:
        dest = os.getcwd()

    filepath = os.path.join(dest, _filename(url))

    with session_line(f"Downloading {url}... ", disable=not verbose):
        urlretrieve(url, filepath)

    return filepath


def _filename(url):
    import os
    from urllib.parse import urlparse

    a = urlparse(url)
    return os.path.basename(a.path)



import click

import limix


@click.command()
@click.pass_context
@click.argument("input-file", type=click.Path(exists=True))
@click.option(
    "--output-file",
    help="Specify the output file path.",
    default=None,
    type=click.Path(exists=False),
)
@click.option(
    "--filetype", help="Specify the file type instead of guessing.", default="guess"
)
@click.option(
    "--verbose/--quiet", "-v/-q", help="Enable or disable verbose mode.", default=True
)
def estimate_kinship(ctx, input_file, output_file, filetype, verbose):
    """Estimate a kinship matrix."""

    if filetype == "guess":
        filetype = limix.io.detect_file_type(input_file)

    if verbose:
        print("Detected file type: {}".format(filetype))

    if filetype == "bgen":
        raise NotImplementedError()
        # G = limix.io.bgen._read_dosage(input_file, verbose=verbose)
    elif filetype == "bed":
        G = limix.io.plink._read_dosage(input_file, verbose=verbose)
    else:
        print("Unknown file type: %s" % input_file)

    K = limix.stats.linear_kinship(G, verbose=verbose)

    if output_file is None:
        output_file = input_file + ".npy"

    output_file, oft = limix.io.detect_file_type(output_file)

    if oft == "npy":
        limix.io.npy.save_kinship(output_file, K, verbose=verbose)
    else:
        print("Unknown output file type: %s" % output_file)



from collections import namedtuple

FetchSpec = namedtuple("FetchSpec", ["filepath", "filetype", "matrix_spec"])


def fetch(target, fetch_spec, verbose=True):
    from .._data import asarray
    from .._data import assert_target, assert_filetype
    from .._data import CONF

    assert_target(target)

    if isinstance(fetch_spec, str):
        fetch_spec = parse_fetch_spec(fetch_spec)

    assert_filetype(fetch_spec.filetype)

    X = _dispatch[target][fetch_spec.filetype](fetch_spec.filepath, verbose=verbose)

    matrix_spec = fetch_spec.matrix_spec
    dims = {d: matrix_spec[d] for d in ["row", "col"] if d in matrix_spec}
    if len(matrix_spec) == 1:
        if not hasattr(X, "dims") or all([d not in CONF["dim_names"] for d in X.dims]):
            dims = _default_dims[target][fetch_spec.filetype]()

    X = asarray(X, target, dims)

    if len(matrix_spec["sel"]) > 0:
        X = X.sel(**matrix_spec["sel"])

    X = asarray(X, target)
    X.name = target

    return X


def _fetch_npy_covariance(filepath, verbose=True):
    from .npy import read

    return read(filepath, verbose=verbose)


def _fetch_bed_genotype(filepath, verbose=True):
    from .plink import read
    from limix._data import asarray

    candidates, samples, G = read(filepath, verbose=verbose)

    assert G.shape == (len(candidates), len(samples))
    G = asarray(G, "genotype", ["candidate", "sample"])
    # Make sure we return the same matrix layout that has been read.
    G = G.transpose("candidate", "sample")

    for colname in samples.columns:
        G.coords[colname] = ("sample", samples[colname].values)

    G.coords[samples.index.name] = ("sample", samples.index)

    for colname in candidates.columns:
        G.coords[colname] = ("candidate", candidates[colname].values)

    G.coords[candidates.index.name] = ("candidate", candidates.index)

    return G


def _fetch_bimbam_trait(filepath, verbose):
    from .bimbam import read_phenotype

    return read_phenotype(filepath, verbose=verbose)


def _fetch_csv_trait(filepath, verbose):
    from .csv import read

    return read(filepath, verbose=verbose)


def _fetch_csv_covariate(filepath, verbose):
    from .csv import read

    return read(filepath, verbose=verbose)


def _default_csv_covariate_dims():
    return {"row": "sample", "col": "covariate"}


def _default_csv_trait_dims():
    return {"row": "trait", "col": "sample"}


_dispatch = {
    "genotype": {"bed": _fetch_bed_genotype},
    "trait": {"bimbam-pheno": _fetch_bimbam_trait, "csv": _fetch_csv_trait},
    "covariance": {"npy": _fetch_npy_covariance},
    "covariate": {"csv": _fetch_csv_covariate},
}

_default_dims = {
    "covariate": {"csv": _default_csv_covariate_dims},
    "trait": {"csv": _default_csv_trait_dims},
}


def parse_fetch_spec(spec):
    import os
    from ._detect import infer_filetype

    drive, spec = os.path.splitdrive(spec)

    ncolons = sum([1 for c in spec if c == ":"])
    if ncolons > 2:
        raise ValueError("Wrong specification syntax: there were more than two colons.")

    spec = spec + ":" * (2 - ncolons)
    filepath, filetype, matrix_spec = spec.split(":")
    filepath = drive + filepath

    spec = {"filepath": filepath, "filetype": filetype, "matrix_spec": matrix_spec}

    if spec["filetype"] == "":
        spec["filetype"] = infer_filetype(spec["filepath"])
    spec["matrix_spec"] = _parse_matrix_spec(spec["matrix_spec"])

    return FetchSpec(**spec)


def _number_or_string(val):
    if "." in val:
        try:
            val = float(val)
            return val
        except ValueError:
            pass

    try:
        val = int(val)
        return val
    except ValueError:
        pass

    enclosed = False
    if val.startswith("'") and val.endswith("'") and len(val) > 1:
        enclosed = True

    if val.startswith("'") and val.endswith("'") and len(val) > 1:
        enclosed = True

    if not enclosed:
        val = '"' + val + '"'

    return val


def _parse_matrix_spec(txt):
    import re

    parts = _split_matrix_spec(txt)
    data = {"sel": {}}
    for p in parts:
        p = p.strip()
        if p.startswith("row"):
            data["row"] = p.split("=")[1]
        elif p.startswith("col"):
            data["col"] = p.split("=")[1]
        else:
            match = re.match(r"(^[^\[]+)\[(.+)\]$", p)
            if match is None:
                raise ValueError("Invalid fetch specification syntax.")
            # TODO: replace eval for something safer
            v = _number_or_string(match.group(2))
            data["sel"].update({match.group(1): eval(v)})

    return data


def _split_matrix_spec(txt):

    brackets = 0
    parts = []
    j = 0
    for i in range(len(txt)):
        if txt[i] == "[":
            brackets += 1
        elif txt[i] == "]":
            brackets -= 1
        elif txt[i] == "," and brackets == 0:
            if j == i:
                raise ValueError("Invalid fetch specification syntax.")
            parts.append(txt[j:i])
            j = i + 1
        if brackets < 0:
            raise ValueError("Invalid fetch specification syntax.")

    if len(txt[j:]) > 0:
        parts.append(txt[j:])

    return parts



def test(verbose=True):
    """
    Run tests to verify this package's integrity.

    Parameters
    ----------
    verbose : bool
        ``True`` to show diagnostic. Defaults to ``True``.

    Returns
    -------
    int
        Exit code: ``0`` for success.
    """
    from .conftest import pandas_format

    pandas_format()

    args = [
        "--doctest-plus",
        "--doctest-plus-rtol=1e-04",
        "--doctest-plus-atol=1e-04",
        "--doctest-modules",
    ]
    if not verbose:
        args += ["--quiet"]

    args += ["--pyargs", __name__.split(".")[0]]

    return __import__("pytest").main(args)



CONF = {
    "likelihoods": set(["normal", "bernoulli", "probit", "binomial", "poisson"]),
    "targets": set(
        [
            "trait",
            "covariate",
            "covariance",
            "genotype",
            "covariate",
            "inter0",
            "inter1",
            "env0",
            "env1",
        ]
    ),
    "filetypes": set(["csv", "bed"]),
    "dim_axis": {
        "sample": 0,
        "trait": 1,
        "candidate": 1,
        "covariate": 1,
        "sample_0": 0,
        "sample_1": 1,
    },
    "dim_names": {"sample", "candidate", "covariate", "trait"},
    "data_synonym": {
        "y": "trait",
        "trait": "y",
        "G": "genotype",
        "genotype": "G",
        "M": "covariate",
        "covariate": "M",
        "K": "covariance",
        "covariance": "K",
    },
    "data_dims": {
        "trait": ["sample", "trait"],
        "genotype": ["sample", "candidate"],
        "covariate": ["sample", "covariate"],
        "covariance": ["sample_0", "sample_1"],
        "inter0": ["sample", "inter"],
        "inter1": ["sample", "inter"],
        "env0": ["sample", "env"],
        "env1": ["sample", "env"],
    },
    "varname_to_target": {
        "y": "trait",
        "M": "covariate",
        "G": "genotype",
        "K": "covariance",
    },
    "target_to_varname": {
        "trait": "y",
        "covariate": "M",
        "genotype": "G",
        "covariance": "K",
    },
}



from limix._bits import unvec
from limix._cache import cache


class MTSimpleModelResult:
    def __init__(self, lik, traits, covariates, lml, beta, beta_se, C0, C1):
        from numpy import asarray, atleast_1d, atleast_2d

        self._lik = lik
        self._traits = asarray(atleast_1d(traits), str)
        self._covariates = asarray(atleast_1d(covariates), str)
        self._lml = float(lml)
        self._beta = atleast_1d(asarray(beta, float).T).T
        self._beta_se = atleast_1d(asarray(beta_se, float).T).T
        self._C0 = atleast_2d(asarray(C0, float))
        self._C1 = atleast_2d(asarray(C1, float))

    @property
    def traits(self):
        return self._traits

    @property
    def likelihood(self):
        return self._lik

    @property
    def lml(self):
        return self._lml

    @property
    def effsizes(self):
        return self._dataframes["effsizes"]

    @property
    def variances(self):
        return self._dataframes["variances"]

    @property
    @cache
    def _dataframes(self):
        from pandas import DataFrame

        effsizes = []
        B = unvec(self._beta, (len(self._covariates), -1))
        Bvar = unvec(self._beta_se, (len(self._covariates), -1))
        for i, trait in enumerate(self._traits):
            for j, c in enumerate(self._covariates):
                effsizes.append([trait, c, B[j, i], Bvar[j, i]])

        columns = ["trait", "covariate", "effsize", "effsize_se"]
        df0 = DataFrame(effsizes, columns=columns)

        variances = []
        for i, trait0 in enumerate(self._traits):
            for j, trait1 in enumerate(self._traits):
                variances.append([trait0, trait1, self._C0[i, j], self._C1[i, j]])

        columns = ["trait0", "trait1", "fore_covariance", "back_covariance"]
        df1 = DataFrame(variances, columns=columns)

        return {"effsizes": df0, "variances": df1}



from .._display import session_line as _session_line


def read(filepath, verbose=True):
    """
    Read NumPy arrays saved in a file.

    Parameters
    ----------
    filepath : str
        File path.
    verbose : bool, optional
        Defaults to ``True``.

    Returns
    -------
    array : array, tuple, dict, etc.
        Data stored in the file.
    """
    from numpy import load

    with _session_line("Reading {}...".format(filepath), disable=not verbose):
        return load(filepath)


def _see(filepath, verbose=True):
    print(read(filepath, verbose=verbose))



from ._core import blue, bold, green, red, width
from ._display import add_title_header, banner, indent, session_block, session_line
from ._exception import print_exc

__all__ = [
    "print_exc",
    "banner",
    "bold",
    "blue",
    "red",
    "green",
    "width",
    "session_block",
    "session_line",
    "add_title_header",
    "indent",
]



from __future__ import division

from brent_search import brent


def boxcox(x):
    r"""Box-Cox transformation for normality conformance.

    It applies the power transformation

    .. math::

        f(x) = \begin{cases}
            \frac{x^{\lambda} - 1}{\lambda}, & \text{if } \lambda > 0; \\
            \log(x), & \text{if } \lambda = 0.
        \end{cases}

    to the provided data, hopefully making it more normal distribution-like.
    The λ parameter is fit by maximum likelihood estimation.

    Parameters
    ----------
    X : array_like
        Data to be transformed.

    Returns
    -------
    boxcox : ndarray
        Box-Cox transformed data.

    Examples
    --------
    .. plot::

        >>> import limix
        >>> import numpy as np
        >>> import scipy.stats as stats
        ...
        >>> np.random.seed(0)
        ...
        >>> x = stats.loggamma.rvs(0.1, size=100)
        >>> y = limix.qc.boxcox(x)
        ...
        >>> plt = limix.plot.get_pyplot()
        ...
        >>> _, (ax1, ax2) = plt.subplots(2, 1)
        >>> _ = stats.probplot(x, dist=stats.norm, plot=ax1)
        >>> _ = stats.probplot(y, dist=stats.norm, plot=ax2)
        >>> plt.tight_layout()
    """
    import dask.array as da
    import numpy as np

    if isinstance(x, da.Array):
        return _boxcox(da, x)
    return _boxcox(np, x)


def _boxcox(lib, x):
    from numpy_sugar import epsilon
    from scipy.stats import boxcox_llf
    from scipy.special import boxcox as bc

    x = lib.asarray(x).astype(float)

    m = x.min()
    if m <= 0:
        m = max(lib.abs(m), epsilon.small)
        x = x + m + m / 2

    lmb = brent(lambda lmb: -boxcox_llf(lmb, x), -5, +5)[0]
    return bc(x, lmb)



import click

import limix


@click.command()
@click.pass_context
@click.argument("filepath")
@click.option(
    "--show_chunks",
    help="Chunks if datasets will be displayed, if available.",
    default="guess",
)
@click.option(
    "--header / --no-header",
    help="Parse header from CSV file. Defaults to false.",
    default=False,
)
@click.option(
    "--verbose/--quiet", "-v/-q", help="Enable or disable verbose mode.", default=True
)
def see(ctx, filepath, show_chunks, header, verbose):
    """
    Show an overview of multiple file types.
    """
    from limix.io._fetch import parse_fetch_spec

    spec = parse_fetch_spec(filepath)
    filetype = spec.filetype
    filepath = spec.filepath

    if filetype == "unknown":
        print("Unknown file type or file path not reachable: `%s`." % filepath)

    # if filetype == "guess":
    #     filepath, filetype = limix.io.detect_file_type(filepath)

    if filetype == "hdf5":
        limix.io.hdf5._see(filepath, show_chunks=show_chunks)

    elif filetype == "csv":
        limix.io.csv._see(filepath, verbose=verbose, header=header)

    elif filetype == "grm.raw":
        limix.io.plink._see_kinship(filepath, verbose)
        limix.plot.show()

    elif filetype == "bed":
        limix.io.plink._see_bed(filepath, verbose)

    elif filetype == "bimbam-pheno":
        limix.io.bimbam._see_phenotype(filepath, verbose)

    elif filetype == "npy":
        limix.io.npy._see(filepath, verbose)

    elif filetype == "image":
        r = limix.plot.image(filepath)
        limix.plot.show()
        return r



def normalise_covariance(K, out=None):
    """
    Variance rescaling of covariance matrix 𝙺.

    Let n be the number of rows (or columns) of 𝙺 and let
    mᵢ be the average of the values in the i-th column.
    Gower rescaling is defined as

    .. math::

        𝙺(n - 1)/(𝚝𝚛𝚊𝚌𝚎(𝙺) - ∑mᵢ).

    Notes
    -----
    The reasoning of the scaling is as follows.
    Let 𝐠 be a vector of n independent samples and let 𝙲 be the Gower's centering
    matrix.
    The unbiased variance estimator is

    .. math::

        v = ∑ (gᵢ-ḡ)²/(n-1) = 𝚝𝚛𝚊𝚌𝚎((𝐠-ḡ𝟏)ᵀ(𝐠-ḡ𝟏))/(n-1) = 𝚝𝚛𝚊𝚌𝚎(𝙲𝐠𝐠ᵀ𝙲)/(n-1)

    Let 𝙺 be the covariance matrix of 𝐠.
    The expectation of the unbiased variance estimator is

    .. math::

        𝐄[v] = 𝚝𝚛𝚊𝚌𝚎(𝙲𝐄[𝐠𝐠ᵀ]𝙲)/(n-1) = 𝚝𝚛𝚊𝚌𝚎(𝙲𝙺𝙲)/(n-1),

    assuming that 𝐄[gᵢ]=0.
    We thus divide 𝙺 by 𝐄[v] to achieve an unbiased normalisation on the random variable
    gᵢ.

    Parameters
    ----------
    K : array_like
        Covariance matrix to be normalised.
    out : array_like, optional
        Result destination. Defaults to ``None``.

    Examples
    --------
    .. doctest::

        >>> from numpy import dot, mean, zeros
        >>> from numpy.random import RandomState
        >>> from limix.qc import normalise_covariance
        >>>
        >>> random = RandomState(0)
        >>> X = random.randn(10, 10)
        >>> K = dot(X, X.T)
        >>> Z = random.multivariate_normal(zeros(10), K, 500)
        >>> print("%.3f" % mean(Z.var(1, ddof=1)))
        9.824
        >>> Kn = normalise_covariance(K)
        >>> Zn = random.multivariate_normal(zeros(10), Kn, 500)
        >>> print("%.3f" % mean(Zn.var(1, ddof=1)))
        1.008

    .. _Dask: https://dask.pydata.org/
    """
    from numpy import asarray
    import dask.array as da
    from pandas import DataFrame
    import xarray as xr

    if isinstance(K, DataFrame):
        K = K.astype(float)
        trace = K.values.trace()
    elif isinstance(K, da.Array):
        trace = da.diag(K).sum()
    elif isinstance(K, xr.DataArray):
        trace = da.diag(K.data).sum()
    else:
        K = asarray(K, float)
        trace = K.trace()

    c = asarray((K.shape[0] - 1) / (trace - K.mean(axis=0).sum()), float)
    if out is None:
        return K * c

    _copyto(out, K)
    _inplace_mult(out, c)

    return out


def _copyto(dst, src):
    from numpy import copyto, ndarray
    import dask.array as da
    from pandas import DataFrame

    if isinstance(dst, DataFrame):
        copyto(dst.values, src)
    elif isinstance(dst, ndarray) and isinstance(src, da.Array):
        copyto(dst, src.compute())
    else:
        copyto(dst, src)


def _inplace_mult(out, c):
    import dask.array as da

    if isinstance(c, da.Array):
        c = c.compute()
    out *= c



class Aligned:
    def __init__(self):
        self._items = []

    def add_item(self, field, value):
        self._items.append([field, value])

    def draw(self):
        from textwrap import TextWrapper

        max_len = max(len(i[0]) for i in self._items)
        max_len += 3

        for i in self._items:
            i[0] = i[0] + " " * (max_len - 2 - len(i[0]))
            i[0] += "= "

        s = " " * max_len

        msg = ""
        for i in self._items:
            wrapper = TextWrapper(initial_indent=i[0], width=88, subsequent_indent=s)
            msg += wrapper.fill(str(i[1])) + "\n"

        return msg



def makedirs(dirpath):
    import os
    import sys

    if sys.version_info >= (3,):
        os.makedirs(dirpath, exist_ok=True)
    else:
        try:
            os.makedirs(dirpath)
        except OSError:
            pass



import dask.array as da
import xarray as xr
from bgen_reader import allele_expectation, compute_dosage, example_files, read_bgen
from numpy import array_equal, ndarray
from numpy.random import RandomState
from numpy.testing import assert_, assert_allclose
from pandas import DataFrame, Series

from limix.qc import compute_maf


def test_compute_dosage():

    with example_files("example.32bits.bgen") as filepath:
        bgen = read_bgen(filepath, verbose=False)
        variant_idx = 2
        e = allele_expectation(bgen, variant_idx)
        dosage = compute_dosage(e)
        assert_allclose(
            dosage[:5],
            [
                0.015502935046214363,
                0.9938354277968955,
                1.9793395833064196,
                0.9956054727070978,
                1.978790270625332,
            ],
        )


def test_compute_maf_numpy():
    random = RandomState(0)
    X = random.randint(0, 3, size=(100, 10))

    maf = compute_maf(X)
    assert_allclose(maf, [0.49, 0.49, 0.445, 0.495, 0.5, 0.45, 0.48, 0.48, 0.47, 0.435])


def test_compute_maf_dataframe():
    random = RandomState(0)

    X = random.randint(0, 3, size=(100, 10))
    columns = [f"snp{i}" for i in range(X.shape[1])]
    maf = compute_maf(DataFrame(X, columns=columns))

    assert_(isinstance(maf, Series))
    assert_(maf.name == "maf")
    assert_(array_equal(maf.index, columns))
    assert_allclose(maf, [0.49, 0.49, 0.445, 0.495, 0.5, 0.45, 0.48, 0.48, 0.47, 0.435])


def test_compute_maf_dask_array():
    random = RandomState(0)

    X = da.from_array(random.randint(0, 3, size=(100, 10)), chunks=2)
    maf = compute_maf(X)

    assert_(isinstance(maf, ndarray))
    assert_allclose(maf, [0.49, 0.49, 0.445, 0.495, 0.5, 0.45, 0.48, 0.48, 0.47, 0.435])


def test_compute_maf_dataarray():
    random = RandomState(0)

    X = random.randint(0, 3, size=(100, 10))
    samples = [f"snp{i}" for i in range(X.shape[0])]
    candidates = [f"snp{i}" for i in range(X.shape[1])]
    X = xr.DataArray(
        X,
        dims=["sample", "candidate"],
        coords={"sample": samples, "candidate": candidates},
    )
    maf = compute_maf(X)

    assert_(isinstance(maf, xr.DataArray))
    assert_(maf.name == "maf")
    assert_(array_equal(maf.candidate, candidates))
    assert_allclose(maf, [0.49, 0.49, 0.445, 0.495, 0.5, 0.45, 0.48, 0.48, 0.47, 0.435])



"""
_bits module
============

Private subpackage for low-level operations.

Modules
-------
dask
deco
pandas
xarray

Functions
---------
get_shape
"""
from . import dask, deco, numpy, pandas, xarray
from .array import cdot, get_shape, unvec, vec

__all__ = [
    "dask",
    "xarray",
    "deco",
    "pandas",
    "get_shape",
    "numpy",
    "vec",
    "unvec",
    "cdot",
]



import click

import limix


@click.command()
@click.pass_context
@click.argument("filepath", type=click.Path(exists=True))
@click.option(
    "--verbose/--quiet", "-v/-q", help="Enable or disable verbose mode.", default=True
)
def extract(ctx, filepath, verbose):
    """Extract a file."""
    limix.sh.extract(filepath, verbose=verbose)



from __future__ import division


def lrt_pvalues(null_lml, alt_lmls, dof=1):
    """
    Compute p-values from likelihood ratios.

    These are likelihood ratio test p-values.

    Parameters
    ----------
    null_lml : float
        Log of the marginal likelihood under the null hypothesis.
    alt_lmls : array_like
        Log of the marginal likelihoods under the alternative hypotheses.
    dof : int
        Degrees of freedom.

    Returns
    -------
    pvalues : ndarray
        P-values.
    """
    from scipy.stats import chi2
    from numpy_sugar import epsilon
    from numpy import asarray, clip, inf

    lrs = clip(-2 * null_lml + 2 * asarray(alt_lmls, float), epsilon.super_tiny, inf)
    pv = chi2(df=dof).sf(lrs)
    return clip(pv, epsilon.super_tiny, 1 - epsilon.tiny)



def get_shape(x):
    from . import dask

    if dask.is_array(x) or dask.is_dataframe(x):
        return _get_dask_shape(x)
    return x.shape


def _get_dask_shape(x):
    import dask.array as da

    return da.compute(*x.shape)


def vec(x):
    from numpy import reshape

    return reshape(x, (-1,) + x.shape[2:], order="F")


def unvec(x, shape):
    from numpy import reshape

    return reshape(x, shape, order="F")


def cdot(A, B):
    """
    𝙰⊙𝙱 = [𝙰₀𝙱₀ ... 𝙰₀𝙱ₙ 𝙰₁𝙱₀ ... 𝙰₁𝙱ₙ ... 𝙰ₘ𝙱ₙ].
    """
    from numpy import tile, repeat

    BB = tile(B, A.shape[1])
    AA = repeat(A, B.shape[1], axis=1)
    return AA * BB



import sys
from time import time

from ._core import blue, bold, pprint, red, width, wrap_text


def banner():
    from limix import __version__

    pyver = sys.version.split("\n")[0].strip()
    return "Running Limix {} using Python {}.".format(__version__, pyver)


def add_title_header(title, df):
    msg = repr(df)
    k = msg.find("\n") - len(title) - 2
    left = ("-" * (k // 2)) + " "
    right = " " + ("-" * (k // 2 + k % 2))
    out = left + title + right + "\n"
    out += msg
    return out


class session_line(object):
    """
    Print the elapsed time after the execution of a block of code.
    """

    def __init__(self, desc="Running... ", disable=False):
        self._disable = disable
        self._tstart = None
        self._desc = desc
        self.elapsed = None

    def __enter__(self):
        self._tstart = time()
        if not self._disable:
            sys.stdout.write(self._desc)
            sys.stdout.flush()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        from humanfriendly import format_timespan
        from limix.__config__ import get_info

        self.elapsed = time() - self._tstart
        fail = exception_type is not None

        if not self._disable:
            if get_info("rich_text") and not get_info("building_doc"):
                # New line, get back to previous line, and advance cursor to the end
                # of the line. This allows us to always get back to the right cursor
                # position, as long as the cursor is still in the correct line.
                print("\n\033[1A\033[{}C".format(len(self._desc)), end="")
            if fail:
                msg = bold(red("failed"))
                msg += " ({}).".format(format_timespan(self.elapsed))
                pprint(msg)
            else:
                print("done (%s)." % format_timespan(self.elapsed))
                sys.stdout.flush()


class session_block(object):
    """Print session block: session start and session end."""

    def __init__(self, session_name, disable=False):
        self._session_name = session_name
        self._start = None
        self._disable = disable

    def __enter__(self):
        self._start = time()
        msg = " {} session starts ".format(self._session_name)
        if not self._disable:
            msg = wrap_text(msg, width())
            pprint(bold(blue(msg)))

    def __exit__(self, exception_type, exception_value, traceback):
        elapsed = time() - self._start
        fail = exception_type is not None

        if fail:
            msg = " {} session fails in {:.2f} seconds "
            color = red
        else:
            msg = " {} session ends in {:.2f} seconds "
            color = blue

        msg = msg.format(self._session_name, elapsed)
        if not self._disable:
            msg = wrap_text(msg, width())
            pprint(bold(color(msg)))


def indent(txt, size=2):
    space = " " * size
    return space + ("\n" + space).join(txt.split("\n"))



from __future__ import division


def indep_pairwise(X, window_size, step_size, threshold, verbose=True):
    """
    Determine pair-wise independent variants.

    Independent variants are defined via squared Pearson correlations between
    pairs of variants inside a sliding window.

    Parameters
    ----------
    X : array_like
        Sample by variants matrix.
    window_size : int
        Number of variants inside each window.
    step_size : int
        Number of variants the sliding window skips.
    threshold : float
        Squared Pearson correlation threshold for independence.
    verbose : bool
        `True` for progress information; `False` otherwise.

    Returns
    -------
    ok : ndarray
        Boolean array defining independent variants

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.qc import indep_pairwise
        >>>
        >>> random = RandomState(0)
        >>> X = random.randn(10, 20)
        >>>
        >>> indep_pairwise(X, 4, 2, 0.5, verbose=False)
        array([ True,  True, False,  True,  True,  True,  True,  True,  True,
                True,  True,  True,  True,  True,  True,  True,  True,  True,
                True,  True])
    """
    from joblib import Parallel, delayed
    from tqdm import tqdm
    from ..threads import get_max_nthreads
    from numpy import ascontiguousarray, logical_not, zeros

    left = 0
    excls = zeros(X.shape[1], dtype=bool)

    if step_size > window_size:
        raise ValueError("Window size has to be smaller than step size.")

    n = (X.shape[1] + step_size) // step_size

    steps = list(range(n))
    cc = get_max_nthreads()

    with tqdm(total=n, desc="Indep. pairwise", disable=not verbose) as pbar:

        while len(steps) > 0:
            i = 0
            right = 0
            delayeds = []
            while i < len(steps):

                step = steps[i]
                left = step * step_size
                if left < right:
                    i += 1
                    continue

                del steps[i]
                right = min(left + window_size, X.shape[1])
                x = ascontiguousarray(X[:, left:right].T)

                delayeds.append(delayed(_func)(x, excls[left:right], threshold))
                if len(delayeds) == cc:
                    Parallel(n_jobs=min(len(delayeds), cc), backend="threading")(
                        delayeds
                    )
                    pbar.update(len(delayeds))
                    delayeds = []

            if len(delayeds) == 0:
                continue

            Parallel(n_jobs=min(len(delayeds), cc), backend="threading")(delayeds)
            pbar.update(len(delayeds))

    return logical_not(excls)


def _row_norms(X):
    from numpy import double, einsum, sqrt

    norms = einsum("ij,ij->i", X, X, dtype=double)
    return sqrt(norms, out=norms)


def _sq_pearson(X):
    from numpy import ascontiguousarray, double, newaxis, zeros
    from scipy.spatial import _distance_wrap

    m = X.shape[0]
    dm = zeros((m * (m - 1)) // 2, dtype=double)

    X2 = X - X.mean(1)[:, newaxis]
    X2 = ascontiguousarray(X2)
    if hasattr(_distance_wrap, "pdist_cosine_wrap"):
        norms = _row_norms(X2)

    X2 = X - X.mean(axis=1, keepdims=True)

    if hasattr(_distance_wrap, "pdist_cosine_wrap"):
        _distance_wrap.pdist_cosine_wrap(X2, dm, norms)
    else:
        _distance_wrap.pdist_cosine_double_wrap(X2, dm)

    return (-dm + 1) ** 2


def _pdist_threshold(mark, dist, thr):
    mark[:] = False
    size = len(mark)

    m = 0
    for i in range(0, size - 1):
        if mark[i]:
            m += size - (i + 1)
            continue

        for j in range(i + 1, size):
            if dist[m] > thr:
                mark[j] = True
            m += 1


def _func(x, excls, threshold):
    from numpy import zeros

    dist = _sq_pearson(x)
    e = zeros(x.shape[0], dtype=bool)
    _pdist_threshold(e, dist, threshold)
    excls |= e



class Pipeline(object):
    def __init__(self, data):
        self._process = []
        self._data = data
        self._layout = _LayoutChange()

    def append(self, process, name, *args, **kwargs):
        self._process.append(
            {"func": process, "name": name, "args": args, "kwargs": kwargs}
        )

    def run(self, verbose=True):
        from limix._data import conform_dataset, CONF

        target = CONF["varname_to_target"]
        self._layout.append(
            "initial", {target[vn]: self._data[vn].shape for vn in self._data.keys()}
        )

        for p in self._process:
            self._data = p["func"](self._data, *p["args"], **p["kwargs"])
            self._data = conform_dataset(**self._data)

            self._layout.append(
                p["name"],
                {
                    target[n]: self._data[n].shape
                    for n, v in self._data.items()
                    if v is not None
                },
            )

            if self._get_samples().size == 0:
                print(self._layout.to_string())
                raise RuntimeError("Exiting early because there is no sample left.")

        if verbose:
            print(self._layout.to_string())

        return self._data

    def _get_samples(self):
        for x in self._data.values():
            if hasattr(x, "sample"):
                return x.sample
        raise RuntimeError("Could not get samples.")


class _LayoutChange(object):
    def __init__(self):
        self._steps = []

    def append(self, name, target_shapes):
        self._steps.append((name, target_shapes))

    def to_string(self):
        from texttable import Texttable

        table = Texttable()
        header = [""]
        shapes = {k: [k] for k in self._steps[0][1].keys()}

        for step in self._steps:
            header.append(step[0])
            for t, shape in step[1].items():
                shapes[t].append(shape)

        table.header(header)

        table.set_cols_dtype(["t"] * len(header))
        table.set_cols_align(["l"] * len(header))
        table.set_deco(Texttable.HEADER)

        for target in self._steps[0][1].keys():
            table.add_row(shapes[target])

        msg = table.draw()

        msg = self._add_caption(msg, "-", "Table: Data layout transformation.")
        return msg

    def _add_caption(self, msg, c, caption):
        n = len(msg.split("\n")[-1])
        msg += "\n" + (c * n)
        msg += "\n" + caption
        return msg



import dask.array as da
from numpy.random import RandomState
from numpy.testing import assert_allclose

from limix.stats import linear_kinship


def test_kinship_estimation():
    random = RandomState(0)
    X = random.randn(30, 40)

    K0 = linear_kinship(X, verbose=False)

    X = da.from_array(X, chunks=(5, 13))
    K1 = linear_kinship(X, verbose=False)
    assert_allclose(K0, K1)



import dask.array as da
from numpy.random import RandomState
from numpy.testing import assert_array_equal

from limix.qc import unique_variants


def test_unique_variants():
    random = RandomState(0)
    X = random.randn(3, 3)
    X[:, 2] = X[:, 0]
    assert_array_equal(unique_variants(X), X[:, 1:])
    X = da.from_array(X, 1)
    assert_array_equal(unique_variants(X), X[:, 1:])



__all__ = ["is_series", "is_dataframe"]


def is_series(a):
    pkg = a.__class__.__module__.split(".")[0]
    name = a.__class__.__name__

    return pkg == "pandas" and name == "Series"


def is_dataframe(a):
    pkg = a.__class__.__module__.split(".")[0]
    name = a.__class__.__name__

    return pkg == "pandas" and name == "DataFrame"



r"""
******************
Plotting utilities
******************

This module is implemented by limix-plot_ package, from which we export
the following functions and classes:

- :py:class:`~limix_plot.ConsensusCurve`
- :py:func:`~limix_plot.box_aspect`
- :py:func:`~limix_plot.image`
- :py:func:`~limix_plot.kinship`
- :py:func:`~limix_plot.load_dataset`
- :py:func:`~limix_plot.manhattan`
- :py:func:`~limix_plot.normal`
- :py:func:`~limix_plot.pca`
- :py:func:`~limix_plot.power`
- :py:func:`~limix_plot.qqplot`

.. _limix-plot: https://limix-plot.readthedocs.io/
"""
from __future__ import absolute_import

from limix_plot import (
    ConsensusCurve,
    box_aspect,
    get_pyplot,
    image,
    kinship,
    load_dataset,
    manhattan,
    normal,
    pca,
    power,
    qqplot,
    show,
)

__all__ = [
    "box_aspect",
    "ConsensusCurve",
    "image",
    "kinship",
    "load_dataset",
    "manhattan",
    "normal",
    "pca",
    "power",
    "qqplot",
    "image",
    "get_pyplot",
    "show",
]



from .._bits.dask import array_shape_reveal


def allele_frequency(X):
    """
    Compute allele frequency from its expectation.

    ``X`` is a matrix whose last dimension correspond to the different set of
    chromosomes. The first dimension represent different variants and the second
    dimension represent the different samples.

    Parameters
    ----------
    X : array_like
        Allele expectations encoded as a variants-by-samples-by-alleles matrix.

    Returns
    -------
    frequenct : ndarray
        Allele frequencies encoded as a variants-by-alleles matrix.
    """
    ploidy = X.shape[-1]
    if X.ndim < 3:
        n = 1
    else:
        n = X.shape[1]
    return X.sum(-2) / (ploidy * n)


def compute_dosage(X, alt=None):
    """
    Compute dosage from allele expectation.

    Parameters
    ----------
    X : array_like
        Allele expectations encoded as a variants-by-samples-by-alleles matrix.
    ref : array_like
        Allele reference of each locus. The allele having the minor allele frequency for
        the provided ``X`` is used as the reference if `None`. Defaults to ``None``.

    Returns
    -------
    dosage : ndarray
        Dosage encoded as a variants-by-samples matrix.

    Example
    -------

    .. doctest::

        >>> from bgen_reader import read_bgen, allele_expectation, example_files
        >>> from bgen_reader import compute_dosage
        >>>
        >>> with example_files("example.32bits.bgen") as filepath:
        ...     bgen = read_bgen(filepath, verbose=False)
        ...     variant_idx = 2
        ...     e = allele_expectation(bgen, variant_idx)
        ...     dosage = compute_dosage(e)
        ...     print(dosage[:5])
        [0.01550294 0.99383543 1.97933958 0.99560547 1.97879027]
        """
    from numpy import asarray

    if alt is None:
        return X[..., -1]
    try:
        return X[alt, :, alt]
    except NotImplementedError:
        alt = asarray(alt, int)
        return asarray(X, float)[alt, :, alt]


def allele_expectation(p, nalleles, ploidy):
    """
    Allele expectation.

    Compute the expectation of each allele from the given probabilities.
    It accepts three shapes of matrices:
    - unidimensional array of probabilities;
    - bidimensional samples-by-alleles probabilities array;
    - and three dimensional variants-by-samples-by-alleles array.

    Parameters
    ----------
    p : array_like
        Allele probabilities.
    nalleles : int
        Number of alleles.
    ploidy : int
        Number of complete sets of chromosomes.

    Returns
    -------
    expectation : ndarray
        Last dimension will contain the expectation of each allele.

    Examples
    --------

    .. doctest::

        >>> from texttable import Texttable
        >>> from bgen_reader import read_bgen, allele_expectation, example_files
        >>>
        >>> sampleid = "sample_005"
        >>> rsid = "RSID_6"
        >>>
        >>> with example_files("example.32bits.bgen") as filepath:
        ...     bgen = read_bgen(filepath, verbose=False)
        ...
        ...     locus = bgen["variants"].query("rsid == '{}'".format(rsid)).index.compute().item()
        ...     sample = bgen["samples"].to_frame().query("id == '{}'".format(sampleid)).index
        ...     sample = sample[0]
        ...
        ...     nalleles = bgen["variants"].loc[locus]["nalleles"]
        ...     ploidy = 2
        ...
        ...     p = bgen["genotype"][locus].compute()["probs"][sample]
        ...     # For unphased genotypes only.
        ...     e = allele_expectation(bgen, locus)[sample]
        ...
        ...     alleles = bgen["variants"].loc[locus]["allele_ids"].compute().item().split(",")
        ...
        ...     tab = Texttable()
        ...
        ...     print(tab.add_rows(
        ...         [
        ...             ["", "AA", "AG", "GG", "E[.]"],
        ...             ["p"] + list(p) + [1.0],
        ...             ["#" + alleles[0], 2, 1, 0, e[0]],
        ...             ["#" + alleles[1], 0, 1, 2, e[1]],
        ...         ]
        ...     ).draw())
        +----+-------+-------+-------+-------+
        |    |  AA   |  AG   |  GG   | E[.]  |
        +====+=======+=======+=======+=======+
        | p  | 0.012 | 0.987 | 0.001 | 1     |
        +----+-------+-------+-------+-------+
        | #A | 2     | 1     | 0     | 1.011 |
        +----+-------+-------+-------+-------+
        | #G | 0     | 1     | 2     | 0.989 |
        +----+-------+-------+-------+-------+
        >>> print("variant: {}".format(rsid))
        variant: RSID_6
        >>> print("sample : {}".format(sampleid))
        sample : sample_005

    Note
    ----
    This function supports unphased genotypes only.
    """
    from numpy import asarray, newaxis

    g = _get_genotypes(ploidy, nalleles)
    c = asarray(_genotypes_to_allele_counts(g), float)
    c = c.T.reshape((1,) * (p.ndim - 1) + (c.shape[1], c.shape[0]))
    p = array_shape_reveal(p)
    return (c * p[..., newaxis, :]).sum(-1)


def _get_genotypes(ploidy, nalleles):
    g = _make_genotypes(ploidy, 1, nalleles)
    g = sorted([list(reversed(i)) for i in g])
    g = [list(reversed(i)) for i in g]
    return g


def _make_genotypes(ploidy, start, end):
    tups = []
    if ploidy == 0:
        return tups
    if ploidy == 1:
        return [[i] for i in range(start, end + 1)]
    for i in range(start, end + 1):
        t = _make_genotypes(ploidy - 1, i, end)
        for ti in t:
            tups += [[i] + ti]
    return tups


def _genotypes_to_allele_counts(genotypes):
    nalleles = genotypes[-1][0]
    counts = []
    for g in genotypes:
        count = [0] * nalleles
        for gi in g:
            count[gi - 1] += 1
        counts.append(count)
    return counts



import sys

from ._display import bold, pprint, red


def print_exc(stack, e):
    """ Print exception with red and bold fonts. """

    print("Traceback (most recent call last):")
    sys.stdout.write("".join(stack))
    errmsg = "{}: {}".format(e.__class__.__name__, str(e))
    pprint(bold(red(errmsg)))



def remove_dependent_cols(X, tol=1.49e-08):
    """
    Remove dependent columns.

    Return a matrix with dependent columns removed.

    Parameters
    ----------
    X : array_like
        Matrix to might have dependent columns.
    tol : float, optional
        Threshold above which columns are considered dependents.

    Returns
    -------
    rank : ndarray
        Full column rank matrix.
    """
    from scipy.linalg import qr
    from numpy import abs, asarray
    from numpy import concatenate, full

    X = asarray(X)
    if X.shape[1] == 0:
        return X

    r = max(X.shape[1] - X.shape[0], 0)
    i = abs(qr(X, mode="r")[0].diagonal()) > tol
    i = concatenate((i, full(r, True, bool)))
    x = X[:, i]
    if x.shape[1] == X.shape[1]:
        return X
    return remove_dependent_cols(x, tol)



def read(filepath, metadata_file=None, sample_file=None, verbose=True):
    """
    Read a given BGEN file.

    Parameters
    ----------
    filepath : str
        A BGEN file path.
    metadata_file : str, optional
        If ``True``, it will try to read the variants metadata from the
        metadata file ``filepath + ".metadata"``. If this is not possible,
        the variants metadata will be read from the BGEN file itself. If
        ``filepath + ".metadata"`` does not exist, it will try to create one
        with the same name to speed up reads. If ``False``, variants metadata
        will be read only from the BGEN file. If a file path is given instead,
        it assumes that the specified metadata file is valid and readable and
        therefore it will read variants metadata from that file only. Defaults
        to ``None``.
    sample_file : str, optional
        A sample file in `GEN format <http://www.stats.ox.ac.uk/~marchini/software/gwas/file_format.html>`_.
        If sample_file is provided, sample IDs are read from this file. Otherwise, it
        reads from the BGEN file itself if present. Defaults to ``None``.
    verbose : bool, optional
        ``True`` to show progress; ``False`` otherwise.

    Returns
    -------
    variants : :class:`pandas.DataFrame`
        Variant position, chromossomes, RSIDs, etc.
    samples : :class:`pandas.DataFrame`
        Sample identifications.
    genotype : :class:`dask.array.Array`
        Array of genotype references.
    X : :class:`dask.array.Array`
        Allele probabilities.

    Note
    ----
    Metadata files can speed up subsequent reads tremendously. But often the user does
    not have write permission for the default metadata file location
    ``filepath + ".metadata"``. We thus provide the
    :func:`limix.io.bgen.create_metadata_file` function for creating one at the
    given path.
    """
    from bgen_reader import read_bgen

    if verbose:
        print("Reading {}...".format(filepath))
    return read_bgen(filepath, metadata_file, sample_file, verbose=verbose)


def _convert_to_dosage(p, nalleles, ploidy):
    """
    Convert probabilities to dosage.

    Parameters
    ----------
    p : array_like
        Allele probabilities.
    nalleles : int
        Number of alleles.
    ploidy : int
        Number of complete sets of chromosomes.

    Returns
    -------
    :class:`numpy.ndarray`
        Dosage matrix.

    Warning
    -------
    This is a new function that needs more testing. Please, report any problem.
    """
    from bgen_reader import convert_to_dosage

    return convert_to_dosage(p, nalleles, ploidy)


def _create_metadata_file(bgen_filepath, metadata_filepath, verbose=True):
    """
    Create variants metadata file.

    Variants metadata file helps speed up subsequent reads of the associated
    BGEN file.

    Parameters
    ----------
    bgen_filepath : str
        BGEN file path.
    metadata_file : str
        Metadata file path.
    verbose : bool, optional
        ``True`` to show progress; ``False`` otherwise.
    """
    from bgen_reader import create_metadata_file

    create_metadata_file(bgen_filepath, metadata_filepath, verbose=True)



import xarray as xr
from numpy.random import RandomState
from numpy.testing import assert_allclose

import limix
from limix._example import file_example


def test_qtl_xarr():
    with file_example("xarr.hdf5") as filepath:
        sample_ids = limix.io.hdf5.fetch(filepath, "/foo/chr1/col_header/sample_ids")

        rsid = dict()
        for i in range(1, 3):
            rsid[i] = limix.io.hdf5.fetch(
                filepath, "/foo/chr{}/row_header/rsid".format(i)
            )

        G = []
        for i in range(1, 3):
            g = xr.open_dataset(filepath, "/foo/chr{}".format(i))["matrix"]
            g = g.rename({g.dims[0]: "snps", g.dims[1]: "samples"})
            g = g.T
            g["samples"] = sample_ids
            g["snps"] = rsid[i]
            g["chr"] = ("snps", [i] * g.shape[1])
            G.append(g)

        G = xr.concat(G, dim="snps")

        K = limix.stats.linear_kinship(G, verbose=False)

    k = [
        [4.316150754626438, -0.12182214897716158],
        [-0.12182214897716158, 0.25268948339191105],
    ]
    assert_allclose(K[:2][:, :2], k)
    random = RandomState(0)
    y = random.randn(10)

    limix.qtl.scan(G, y, "normal", K, verbose=False)

    G = G.rename({"samples": "sample"})
    limix.qtl.scan(G, y, "normal", K, verbose=False)



from glimix_core.cov import (
    EyeCov,
    FreeFormCov,
    GivenCov,
    Kron2SumCov,
    LinearCov,
    LRFreeFormCov,
    SumCov,
)

__all__ = [
    "EyeCov",
    "FreeFormCov",
    "GivenCov",
    "LinearCov",
    "SumCov",
    "LRFreeFormCov",
    "Kron2SumCov",
]



from numpy.testing import assert_allclose, assert_equal

from limix import file_example, io
from limix.sh import extract


def test_io_hdf5():
    with file_example("data.h5.bz2") as filepath:
        filepath = extract(filepath, verbose=False)
        with io.hdf5.fetcher(filepath) as df:
            X = df.fetch("/genotype/chrom20/SNP")
            assert_allclose(X[0, 0], [1.0])


def test_io_csv():
    with file_example("pheno.csv") as filepath:
        assert_equal(io.csv.read(filepath, verbose=False)["attr1"][0], "string")


def test_io_gen():
    with file_example(["example.gen", "example.sample"]) as filepaths:
        df = io.gen.read(filepaths[0][:-4], verbose=False)
        assert_equal(df["sample"]["sample_id"][1], "1A1")
        assert_equal(df["sample"]["age"][0], 4)
        assert_allclose(df["genotype"]["1A4"]["AB"][0], 0.0207)



from __future__ import division


def compute_maf(X):
    r"""Compute minor allele frequencies.

    It assumes that ``X`` encodes 0, 1, and 2 representing the number
    of alleles (or dosage), or ``NaN`` to represent missing values.

    Parameters
    ----------
    X : array_like
        Genotype matrix.

    Returns
    -------
    array_like
        Minor allele frequencies.

    Examples
    --------
    .. doctest::

        >>> from numpy.random import RandomState
        >>> from limix.qc import compute_maf
        >>>
        >>> random = RandomState(0)
        >>> X = random.randint(0, 3, size=(100, 10))
        >>>
        >>> print(compute_maf(X)) # doctest: +FLOAT_CMP
        [0.49  0.49  0.445 0.495 0.5   0.45  0.48  0.48  0.47  0.435]
    """
    import dask.array as da
    import xarray as xr
    from pandas import DataFrame
    from numpy import isnan, logical_not, minimum, nansum

    if isinstance(X, da.Array):
        s0 = da.nansum(X, axis=0).compute()
        denom = 2 * (X.shape[0] - da.isnan(X).sum(axis=0)).compute()
    elif isinstance(X, DataFrame):
        s0 = X.sum(axis=0, skipna=True)
        denom = 2 * logical_not(X.isna()).sum(axis=0)
    elif isinstance(X, xr.DataArray):
        if "sample" in X.dims:
            kwargs = {"dim": "sample"}
        else:
            kwargs = {"axis": 0}
        s0 = X.sum(skipna=True, **kwargs)
        denom = 2 * logical_not(isnan(X)).sum(**kwargs)
    else:
        s0 = nansum(X, axis=0)
        denom = 2 * logical_not(isnan(X)).sum(axis=0)

    s0 = s0 / denom
    s1 = 1 - s0
    maf = minimum(s0, s1)

    if hasattr(maf, "name"):
        maf.name = "maf"

    return maf



def get_info(name):
    import os

    info = {
        "interactive": _interactive_session(),
        "rich_text": _in_terminal() or _in_ipython_frontend(),
        "building_doc": os.environ.get("LIMIX_DOC", "False") == "True",
    }
    return info[name]


def _interactive_session():
    """
    Check if we're running in an interactive shell.

    Returns
    -------
    ``True`` if running under python/ipython interactive shell; ``False`` otherwise.
    """

    def check_main():
        import __main__ as main

        return not hasattr(main, "__file__")

    try:
        return hasattr(globals()["__builtins__"], "__IPYTHON__") or check_main()
    except Exception:
        return check_main()


def _in_ipython_frontend():
    """
    Check if we're inside an an IPython zmq frontend.
    """
    if "get_ipython" not in globals():
        return False
    ip = globals()["get_ipython"]()
    return "zmq" in str(type(ip)).lower()


def _in_terminal():
    """
    Detect if Python is running in a terminal.

    Returns
    -------
    bool
        ``True`` if Python is running in a terminal; ``False`` otherwise.
    """
    # Assume standard Python interpreter in a terminal.
    if "get_ipython" not in globals():
        return True
    ip = globals()["get_ipython"]()
    # IPython as a Jupyter kernel.
    if hasattr(ip, "kernel"):
        return False
    return True



import click

import limix


@click.command()
@click.pass_context
@click.argument("url")
@click.option(
    "--dest", help="Destination path.", default=None, type=click.Path(exists=True)
)
@click.option(
    "--verbose/--quiet", "-v/-q", help="Enable or disable verbose mode.", default=True
)
def download(ctx, url, dest, verbose):
    """Download file from the specified URL."""
    limix.sh.download(url, dest=dest, verbose=verbose)



class Table:
    def __init__(self, columns, index):
        self._columns = list(columns)
        self._index = list(index)
        self._values = []

    def add_column(self, values):
        from numpy import atleast_1d

        v = atleast_1d(values).ravel().tolist()
        self._values.append(v)

    def draw(self):
        from texttable import Texttable

        table = Texttable(max_width=88)
        table.set_deco(Texttable.HEADER)
        table.set_chars(["", "", "", "-"])
        ncols = len(self._values)
        table.set_cols_dtype(["t"] + ["e"] * ncols)
        table.set_cols_align(["l"] + ["r"] * ncols)

        rows = [[""] + self._columns]
        for i, r in enumerate(zip(*self._values)):
            rows.append([self._index[i]] + list(r))

        table.add_rows(rows)

        return table.draw()



from numpy.random import RandomState
from numpy.testing import assert_, assert_array_equal
from xarray import DataArray

from limix._data import conform_dataset


def test_data_conform_dataset():
    random = RandomState(0)

    y = random.randn(5)
    G = random.randn(5, 2)

    data = conform_dataset(y, G=G)
    y = data["y"]
    G = data["G"]
    assert_(isinstance(y, DataArray))
    assert_(isinstance(G, DataArray))
    assert_array_equal(y.shape, (5, 1))
    assert_array_equal(G.shape, (5, 2))



def read_limix(filepath):
    """
    Read the HDF5 limix file format.

    Parameters
    ----------
    filepath : str
        File path.

    Returns
    -------
    dict
        Phenotype and genotype.
    """
    Y = fetch(filepath, "phenotype/matrix")
    rows = _read_attrs(filepath, "phenotype/row_header")
    cols = _read_attrs(filepath, "phenotype/col_header")
    Y = _convert_headers(Y, rows, cols, "outcome")
    Y = Y.rename(sample_ID="sample")
    Y.name = "phenotype"

    G = fetch(filepath, "genotype/matrix")
    rows = _read_attrs(filepath, "genotype/row_header")
    cols = _read_attrs(filepath, "genotype/col_header")
    G = _convert_headers(G, rows, cols, "candidate")
    G = G.rename(sample_ID="sample")
    G.name = "genotype"

    return {"phenotype": Y, "genotype": G}


class fetcher(object):
    """
    Fetch datasets from HDF5 files.

    Parameters
    ----------
    filename : str
        Filename to an HDF5 file.

    Examples
    --------
    .. doctest::

        >>> from limix.io import hdf5
        >>> from limix import file_example
        >>> from limix.sh import extract
        >>>
        >>> with file_example("data.h5.bz2") as filepath:
        ...     filepath = extract(filepath, verbose=False)
        ...     with hdf5.fetcher(filepath) as df:
        ...         X = df.fetch('/group/dataset')
        ...         print('%.4f' % X[0, 0].compute())
        -0.0453
    """

    def __init__(self, filename):
        self._filename = filename

    def __enter__(self):
        import h5py

        self._f = h5py.File(self._filename, "r")
        return self

    def fetch(self, data_path):
        r"""Fetch a HDF5 dataset.

        Parameters
        ----------
        data_path : str
            Path to a dataset.

        Returns
        -------
        X : dask array
        """
        from dask.array import from_array

        data = self._f[data_path]
        if data.chunks is None:
            chunks = data.shape
        else:
            chunks = data.chunks
        return from_array(data, chunks=chunks)

    def __exit__(self, *args):
        self._f.close()


def fetch(fp, path):
    """
    Fetches an array from hdf5 file.

    Parameters
    ----------
    fp : str
        HDF5 file path.
    path : str
        Path inside the HDF5 file.

    Returns
    -------
    numpy.ndarray
        Array read from the HDF5 dataset.
    """
    import h5py

    with h5py.File(fp, "r") as f:
        return f[path][:]


def _see(f_or_filepath, root_name="/", show_chunks=False):
    """
    Shows a human-friendly tree representation of the contents of a hdf5 file.

    Parameters
    ----------
    file_or_filepath : str, file
        HDF5 file path or a handler to an open one.
    root_name : str, optional
        Group path to be the tree root. Defaults to ``"/"``.
    show_chunks : bool, optional
        ``True`` to show chunks; ``False`` otherwise. Defaults to ``False``.

    Returns
    -------
    str
        String representation of the tree.
    """
    import h5py

    if isinstance(f_or_filepath, str):
        with h5py.File(f_or_filepath, "r") as f:
            _tree(f, root_name, show_chunks)
    else:
        _tree(f_or_filepath, root_name, show_chunks)


def _findnth(haystack, needle, n):
    parts = haystack.split(needle, n + 1)
    if len(parts) <= n + 1:
        return -1
    return len(haystack) - len(parts[-1]) - len(needle)


def _visititems(root, func, level=0, prefix=""):
    if root.name != "/":
        name = root.name
        eman = name[::-1]
        i1 = _findnth(eman, "/", level)
        name = "/" + eman[:i1][::-1]
        func(prefix + name, root)
    if not hasattr(root, "keys"):
        return
    for k in root.keys():
        if root.file == root[k].file:
            _visititems(root[k], func, level + 1, prefix)
        else:
            _visititems(root[k], func, 0, prefix + root.name)


def _tree(f, root_name="/", show_chunks=False):
    import h5py

    _names = []

    def get_names(name, obj):
        if isinstance(obj, h5py.Dataset):
            dtype = str(obj.dtype)
            shape = str(obj.shape)
            if show_chunks:
                chunks = str(obj.chunks)
                _names.append("%s [%s, %s, %s]" % (name[1:], dtype, shape, chunks))
            else:
                _names.append("%s [%s, %s]" % (name[1:], dtype, shape))
        else:
            _names.append(name[1:])

    _visititems(f, get_names)

    class Node(object):
        def __init__(self, name, children):
            self.name = name
            self.children = children

        def __str__(self):
            return self.name

    root = Node(root_name, dict())

    def add_to_node(node, ns):
        if len(ns) == 0:
            return
        if ns[0] not in node.children:
            node.children[ns[0]] = Node(ns[0], dict())
        add_to_node(node.children[ns[0]], ns[1:])

    _names = sorted(_names)
    for n in _names:
        ns = n.split("/")
        add_to_node(root, ns)

    def child_iter(node):
        from numpy import argsort, asarray

        keys = list(node.children.keys())
        indices = argsort(keys)
        indices = asarray(indices)
        vals = list(node.children.values())
        return list(asarray(vals)[indices])

    import asciitree

    print(asciitree.draw_tree(root, child_iter))


def _read_attrs(filepath, path):
    from pandas import DataFrame
    import h5py

    with h5py.File(filepath, "r") as f:

        h = dict()
        for attr in f[path].keys():
            h[attr] = f[path + "/" + attr][()]

            if h[attr].dtype.kind == "S":
                h[attr] = h[attr].astype("U")

        return DataFrame.from_dict(h)


def _convert_headers(X, rows, cols, colname):
    from xarray import DataArray

    coords = {k: ("sample", rows[k]) for k in rows.keys()}
    coords.update({k: (colname, cols[k]) for k in cols.keys()})
    return DataArray(X, dims=["sample", colname], coords=coords)



def filehash(filepath):
    """
    Compute sha256 from a given file.

    Parameters
    ----------
    filepath : str
        File path.

    Returns
    -------
    sha256 : str
        Sha256 of a given file.
    """
    import hashlib

    BUF_SIZE = 65536
    sha256 = hashlib.sha256()

    with open(filepath, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)

    return sha256.hexdigest()



from __future__ import division

import logging


def _get_jit():
    try:
        from numba import jit
    except ImportError:

        def jit(x, *args, **kwargs):
            return x

    return jit


def _get_walk_left():
    jit = _get_jit()

    @jit(cache=True)
    def _walk_left(pos, c, dist):
        step = 0
        middle = pos[c]
        i = c
        while i > 0 and step < dist:
            i -= 1
            step = middle - pos[i]
        if step > dist:
            i += 1
        return i

    return _walk_left


def _get_walk_right():
    jit = _get_jit()

    @jit(cache=True)
    def _walk_right(pos, c, dist):
        step = 0
        middle = pos[c]
        i = c
        n = len(pos)
        while i < n - 1 and step < dist:
            i += 1
            step = pos[i] - middle
        if step > dist:
            i -= 1
        return i

    return _walk_right


def roc_curve(multi_score, method, max_fpr=0.05):
    from numpy import arange, empty_like, mean, std, sqrt

    max_fpr = float(max_fpr)
    fprs = arange(0.0, max_fpr, step=0.001)
    tprs = empty_like(fprs)
    tprs_stde = empty_like(fprs)
    for (i, fpr) in enumerate(fprs):
        tprs_ = multi_score.get_tprs(method, fpr=fpr, approach="rank")
        tprs[i] = mean(tprs_)
        tprs_stde[i] = std(tprs_) / sqrt(len(tprs_))
    return (fprs, tprs, tprs_stde)


# TODO: convert to numpy style
def confusion_matrix(df, wsize=50000):
    """Provide a couple of scores based on the idea of windows around
       genetic markers.

       :param causals: Indices defining the causal markers.
       :param pos: Within-chromossome base-pair position of each candidate
                   marker, in crescent order.
    """
    from numpy import argsort, asarray, concatenate, where

    logger = logging.getLogger(__name__)
    wsize = int(wsize)

    if "chrom" not in df:
        df = df.assign(chrom=["1"] * len(df))

    df.sort_values(by=["chrom", "pos"], inplace=True)

    chromids = df["chrom"].unique()

    offset = 0
    idx_offset = 0
    pos = []
    causal = []
    pv = []
    for cid in sorted(chromids):
        df0 = df.query("chrom=='%s'" % cid)

        pos.append(offset + asarray(df0["pos"], float))
        pv.append(asarray(df0["pv"], float))
        offset += pos[-1][-1] + wsize // 2 + 1

        if df0["causal"].sum() > 0:
            causal.append(idx_offset + where(df0["causal"])[0])
            idx_offset += len(df0)

    pos = concatenate(pos)
    pv = concatenate(pv)
    causal = concatenate(causal)
    causal = asarray(causal, int)

    total_size = pos[-1] - pos[0]
    if wsize > 0.1 * total_size:
        perc = wsize // total_size * 100
        logger.warn(
            "The window size is %d%% of the total candidate" + " region.", int(perc)
        )

    ld_causal_markers = set()
    for _, c in enumerate(causal):
        if wsize == 1:
            right = left = pos[c]
        else:
            left = _get_walk_left()(pos, c, wsize // 2)
            right = _get_walk_right()(pos, c, wsize // 2)
        for i in range(left, right + 1):
            ld_causal_markers.add(i)

    P = len(ld_causal_markers)
    N = len(pos) - P

    ld_causal_markers = list(ld_causal_markers)

    logger.info("Found %d positive and %d negative markers.", P, N)

    return ConfusionMatrix(P, N, ld_causal_markers, argsort(pv))


def getter(func):
    class ItemGetter(object):
        def __getitem__(self, i):
            return func(i)

        def __lt__(self, other):
            from numpy import s_

            return func(s_[:]) < other

        def __le__(self, other):
            from numpy import s_

            return func(s_[:]) <= other

        def __gt__(self, other):
            from numpy import s_

            return func(s_[:]) > other

        def __ge__(self, other):
            from numpy import s_

            return func(s_[:]) >= other

        def __eq__(self, other):
            from numpy import s_

            return func(s_[:]) == other

    return ItemGetter()


# TODO: document it
class ConfusionMatrix(object):
    def __init__(self, P, N, true_set, idx_rank):
        from numpy import empty, asarray, searchsorted

        self._TP = empty(P + N + 1, dtype=int)
        self._FP = empty(P + N + 1, dtype=int)
        if len(idx_rank) != P + N:
            raise ValueError(
                "Rank indices array has to have length equal" + " to ``P + N``."
            )

        true_set = asarray(true_set, int)
        true_set.sort()

        idx_rank = asarray(idx_rank, int)

        ins_pos = searchsorted(true_set, idx_rank)
        _confusion_matrix_tp_fp(P + N, ins_pos, true_set, idx_rank, self._TP, self._FP)
        self._N = N
        self._P = P

    @property
    def TP(self):
        return getter(lambda i: self._TP[i])

    @property
    def FP(self):
        return getter(lambda i: self._FP[i])

    @property
    def TN(self):
        return getter(lambda i: self._N - self.FP[i])

    @property
    def FN(self):
        return getter(lambda i: self._P - self.TP[i])

    @property
    def sensitivity(self):
        """ Sensitivity (also known as true positive rate.)
        """
        return getter(lambda i: self.TP[i] / self._P)

    @property
    def tpr(self):
        return self.sensitivity

    @property
    def recall(self):
        return self.sensitivity

    @property
    def specifity(self):
        """ Specifity (also known as true negative rate.)
        """
        return getter(lambda i: self.TN[i] / self._N)

    @property
    def precision(self):
        from numpy import nan

        return getter(
            lambda i: nan if i == 0 else self.TP[i] / (self.TP[i] + self.FP[i])
        )

    @property
    def npv(self):
        """ Negative predictive value.
        """
        return getter(lambda i: self.TN[i] / (self.TN[i] + self.FN[i]))

    @property
    def fallout(self):
        """ Fall-out (also known as false positive rate.)
        """
        return getter(lambda i: 1 - self.specifity[i])

    @property
    def fpr(self):
        return self.fallout

    @property
    def fnr(self):
        """ False negative rate.
        """
        return getter(lambda i: 1 - self.sensitivity[i])

    @property
    def fdr(self):
        """ False discovery rate.
        """
        return getter(lambda i: 1 - self.precision[i])

    @property
    def accuracy(self):
        """ Accuracy.
        """
        return getter(lambda i: (self.TP[i] + self.TN[i]) / (self._N + self._P))

    @property
    def f1score(self):
        """ F1 score (harmonic mean of precision and sensitivity).
        """

        def denominator(i):
            return 2 * self.TP[i] + self.FP[i] + self.FN[i]

        return getter(lambda i: 2 * self.TP[i] / denominator(i))

    def roc(self):
        from numpy import argsort

        tpr = self.tpr[1:]
        fpr = self.fpr[1:]

        idx = argsort(fpr)
        fpr = fpr[idx]
        tpr = tpr[idx]

        return (fpr, tpr)


def auc(fpr, tpr):
    left = fpr[0]
    area = 0.0
    for i in range(1, len(fpr)):
        width = fpr[i] - left
        area += width * tpr[i - 1]
        left = fpr[i]
    area += (1 - left) * tpr[-1]
    return area


def _confusion_matrix_tp_fp(n, ins_pos, true_set, idx_rank, TP, FP):
    TP[0] = 0
    FP[0] = 0
    i = 0
    while i < n:
        FP[i + 1] = FP[i]
        TP[i + 1] = TP[i]

        j = ins_pos[i]
        if j == len(true_set) or true_set[j] != idx_rank[i]:
            FP[i + 1] += 1
        else:
            TP[i + 1] += 1
        i += 1



r"""Retrieve information about terminal/frontend.

Acknowledgment
--------------
- Pandas Python package for providing code for such a functionality.
"""
import re
import warnings

from .._config import config

_tags = ["bold", "green", "blue", "red"]
_color = {"blue": "#0C68C7", "green": "#19CB00", "red": "#FF3534"}


def pprint(txt):
    """ Print rich text. """
    try:
        from IPython.display import display

        display(_RichText(txt))
    except Exception:
        print(_RichText(txt))


def add_title_header(title, df):
    msg = repr(df)
    k = msg.find("\n") - len(title) - 2
    left = ("-" * (k // 2)) + " "
    right = " " + ("-" * (k // 2 + k % 2))
    out = left + title + right + "\n"
    out += msg
    return out


def format_richtext(txt):
    """Format rich text."""
    return _RichText(txt)


def display(objs):
    """Frontend-aware object display."""
    try:
        from IPython.display import display

        display(objs)
    except Exception:
        print(objs)


def bold(txt):
    """Bold font."""
    return "[bold]" + txt + "[/bold]"


def green(txt):
    """Green color font."""
    return "[green]" + txt + "[/green]"


def blue(txt):
    """Blue color font."""
    return "[blue]" + txt + "[/blue]"


def red(txt):
    """Red color font."""
    return "[red]" + txt + "[/red]"


def width():
    """Display number of columns."""
    if _in_interactive_session():
        if _in_ipython_frontend():
            return config["display.fallback_width"]
    try:
        term = _get_terminal()
        if term is not None:
            return term.width
    except ValueError as e:
        warnings.warn(e)

    return config["display.fallback_width"]


def wrap_text(msg, width, sym="="):
    if width is None:
        width = 79
    width -= len(msg)
    if width <= 0:
        return msg
    left = width // 2
    right = width - left
    return (sym * left) + msg + (sym * right)


def _in_ipython_frontend():
    """Check if we're inside an an IPython zmq frontend."""
    try:
        ip = get_ipython()  # noqa
        return "zmq" in str(type(ip)).lower()
    except Exception:
        pass
    return False


def _is_terminal():
    """Detect if Python is running in a terminal.

    Returns
    -------
    bool
        ``True`` if Python is running in a terminal; ``False`` otherwise.
    """
    try:
        ip = get_ipython()
    except NameError:  # assume standard Python interpreter in a terminal
        return True
    else:
        if hasattr(ip, "kernel"):  # IPython as a Jupyter kernel
            return False
        else:  # IPython in a terminal
            return True


def _in_interactive_session():
    """Check if we're running in an interactive shell.

    Returns
    -------
    ``True`` if running under python/ipython interactive shell; ``False`` otherwise.
    """

    def check_main():
        import __main__ as main

        return not hasattr(main, "__file__")

    try:
        return __IPYTHON__ or check_main()  # noqa
    except Exception:
        return check_main()


def _get_terminal():
    try:
        from blessings import Terminal
    except ImportError:
        return None

    try:
        term = Terminal()
    except ValueError:
        return None
    return term


def _compile(tag):
    expr = r"\[{tag}\](.*?)\[\/{tag}\]".format(tag=tag)
    return re.compile(expr, re.MULTILINE | re.DOTALL)


class _RichText(object):
    def __init__(self, text):
        self._text = text

    def __repr__(self):
        if _is_terminal() or _in_ipython_frontend():
            return _terminal_format(self._text)
        return _plain_format(self._text)

    def _repr_html_(self):
        txt = self._text

        for tag in _tags:
            r = _compile(tag)
            if tag == "bold":
                txt = r.sub("<b>\\1</b>".format(tag), txt)
            else:
                expr = "<span style='color:{}'>\\1</span>".format(_color[tag])
                txt = r.sub(expr, txt)

        return "<pre>{}</pre>".format(txt)


def _terminal_format(txt):
    term = _get_terminal()

    for tag in _tags:
        r = _compile(tag)
        if term is None:
            by = r"\1"
        else:
            by = getattr(term, tag)(r"\1")
        txt = r.sub(by, txt)

    return txt


def _plain_format(txt):

    for tag in _tags:
        r = _compile(tag)
        txt = r.sub("\\1", txt)

    return txt



import click

from .scan import scan


# @click.command(cls=OrderedCommand)
@click.group()
@click.pass_context
# @click.option(
#     "--model",
#     help=("Specify the statistical model to perform the scan."),
#     default="single-trait",
#     type=click.Choice(["single-trait", "struct-lmm"]),
# )
# @click.option(
#     "--lik",
#     help=(
#         "Specify the type of likelihood that will described"
#         " the residual error distribution."
#     ),
#     default="normal",
# )
def qtl(ctx):
    """ Perform genome-wide association scan. """
    pass


qtl.add_command(scan)



import pytest
from numpy import asarray, nan
from numpy.random import RandomState

from limix.qc import mean_impute

from .util import (
    assert_dask_array,
    assert_dask_dataframe,
    assert_ndarray_1d,
    assert_ndarray_2d,
    assert_pandas_dataframe,
    assert_pandas_series,
    assert_xarray_dataarray,
)


@pytest.fixture
def data2d():
    random = RandomState(0)
    X = random.randn(3, 4)
    X[0, 0] = nan

    R = asarray(
        [
            [
                0.882_169_569_178_204_8,
                0.400_157_208_367_223_3,
                0.978_737_984_105_739_2,
                2.240_893_199_201_458,
            ],
            [
                1.867_557_990_149_967_5,
                -0.977_277_879_876_411,
                0.950_088_417_525_589_4,
                -0.151_357_208_297_697_9,
            ],
            [
                -0.103_218_851_793_557_84,
                0.410_598_501_938_372_33,
                0.144_043_571_160_878,
                1.454_273_506_962_975,
            ],
        ]
    )
    Rt = asarray(
        [
            [
                1.206_596_130_558_14,
                0.400_157_208_367_223_3,
                0.978_737_984_105_739_2,
                2.240_893_199_201_458,
            ],
            [
                1.867_557_990_149_967_5,
                -0.977_277_879_876_411,
                0.950_088_417_525_589_4,
                -0.151_357_208_297_697_9,
            ],
            [
                -0.103_218_851_793_557_84,
                0.410_598_501_938_372_33,
                0.144_043_571_160_878,
                1.454_273_506_962_975,
            ],
        ]
    )

    samples = [f"sample{i}" for i in range(X.shape[0])]
    candidates = [f"snp{i}" for i in range(X.shape[1])]

    return {"X": X, "R": R, "Rt": Rt, "samples": samples, "candidates": candidates}


@pytest.fixture
def data1d():
    X = asarray([nan, 0.400_157_208_367_223_3, 0.978_737_984_105_739_2])
    R = asarray(
        [0.689_447_596_236_481_2, 0.400_157_208_367_223_3, 0.978_737_984_105_739_2]
    )
    Rt = R.copy()
    Rt[0] = nan
    samples = [f"sample{i}" for i in range(X.shape[0])]
    return {"X": X, "R": R, "Rt": Rt, "samples": samples, "candidates": None}


def test_impute_ndarray_1d(data1d):
    assert_ndarray_1d(mean_impute, data1d)


def test_impute_ndarray_2d(data2d):
    assert_ndarray_2d(mean_impute, data2d)


def test_impute_pandas_series(data1d):
    assert_pandas_series(mean_impute, data1d)


def test_impute_pandas_dataframe(data2d):
    assert_pandas_dataframe(mean_impute, data2d)


def test_impute_dask_array(data2d):
    assert_dask_array(mean_impute, data2d)


def test_impute_dask_dataframe(data2d):
    assert_dask_dataframe(mean_impute, data2d)


def test_impute_xarray_dataarray(data2d):
    assert_xarray_dataarray(mean_impute, data2d)



import sys

from limix._display import session_line

from .._data import asarray as _asarray, conform_dataset, normalize_likelihood
from .._display import session_block
from ._assert import assert_finite
from ._result import MTScanResultFactory, STScanResultFactory


def scan(
    G, Y, lik="normal", K=None, M=None, idx=None, A=None, A0=None, A1=None, verbose=True
):
    """
    Multi-trait association and interaction testing via linear mixed models.

    Let n, c, and p be the number of samples, covariates, and traits, respectively.
    The outcome variable Y is a n×p matrix distributed according to ::

        vec(Y) ~ N((A ⊗ M) vec(𝚨), K₀ = C₀ ⊗ K + C₁ ⊗ I) under H₀.

    A and M are design matrices of dimensions p×p and n×c provided by the user,
    where X is the usual matrix of covariates commonly used in single-trait models.
    𝚨 is a c×p matrix of fixed-effect sizes per trait.
    C₀ and C₁ are both symmetric matrices of dimensions p×p, for which C₁ is
    guaranteed by our implementation to be of full rank.
    The parameters of the H₀ model are the matrices 𝚨, C₀, and C₁.

    The additional models H₁ and H₂ are define as ::

        vec(Y) ~ N((A ⊗ M) vec(𝚨) + (A₀ ⊗ Gᵢ) vec(𝚩₀), s⋅K₀)

    and ::

        vec(Y) ~ N((A ⊗ M) vec(𝚨) + (A₀ ⊗ Gᵢ) vec(𝚩₀) + (A₁ ⊗ Gᵢ) vec(𝚩₁), s⋅K₀)

    It performs likelihood-ratio tests for the following cases, where the first
    hypothesis is the null one while the second hypothesis is the alternative one:
    - H₀ vs H₁: testing for vec(𝚩₀) ≠ 𝟎 while vec(𝚩₁) = 𝟎
    - H₀ vs H₂: testing for [vec(𝚩₀) vec(𝚩₁)] ≠ 𝟎
    - H₁ vs H₂: testing for vec(𝚩₁) ≠ 𝟎

    It supports generalized linear mixed models (GLMM) when a single trait is used.
    In this case, the following likelihoods are implemented:
    - Bernoulli
    - Probit
    - Binomial
    - Poisson

    Formally, let p(𝜇) be one of the supported probability distributions where 𝜇 is
    its mean. The H₀ model is defined as follows::

        yᵢ ∼ p(𝜇ᵢ=g(zᵢ)) for 𝐳 ∼ 𝓝(..., ...).

    g(⋅) is the corresponding canonical link function for the Bernoulli, Binomial, and
    Poisson likelihoods. The Probit likelihood, on the other hand, is a Bernoulli
    likelihood with probit link function.

    Parameters
    ----------
    G : n×m array_like
        Genetic candidates.
    Y : n×p array_like
        Rows are samples and columns are phenotypes.
    lik : tuple, "normal", "bernoulli", "probit", "binomial", "poisson"
        Sample likelihood describing the residual distribution.
        Either a tuple or a string specifying the likelihood is required. The Normal,
        Bernoulli, Probit, and Poisson likelihoods can be selected by providing a
        string. Binomial likelihood on the other hand requires a tuple because of the
        number of trials: ``("binomial", array_like)``. Defaults to ``"normal"``.
    K : n×n array_like
        Sample covariance, often the so-called kinship matrix.
    M : n×c array_like
        Covariates matrix.
    idx : list
        List of candidate indices that defines the set of candidates to be used in the
        tests.
    A : p×p array_like
        Symmetric trait-by-trait design matrix.
    A0 : p×p₀ array_like
        Matrix A₀, possibility a non-symmetric one.
    A1 : p×p₁ array_like
        Matrix A₁, possibility a non-symmetric one.
    verbose : bool, optional
        ``True`` to display progress and summary; ``False`` otherwise.

    Returns
    -------
    result : :class:`limix.qtl._result.STScanResult`, :class:`limix.qtl._result.MTScanResult`
        P-values, log of marginal likelihoods, effect sizes, and associated statistics.

    Examples
    --------
    .. doctest::

        >>> from limix.qtl import scan
        >>> from numpy import reshape, kron, eye
        >>> from numpy import concatenate
        >>> from numpy.random import RandomState
        >>> import scipy.stats as st
        >>> from limix.qc import normalise_covariance
        >>>
        >>> def vec(x):
        ...     return reshape(x, (-1,) + x.shape[2:], order="F")
        >>>
        >>> def unvec(x, shape):
        ...     return reshape(x, shape, order="F")
        >>>
        >>> random = RandomState(0)
        >>> n = 30
        >>> ntraits = 2
        >>> ncovariates = 3
        >>>
        >>> A = random.randn(ntraits, ntraits)
        >>> A = A @ A.T
        >>> M = random.randn(n, ncovariates)
        >>>
        >>> C0 = random.randn(ntraits, ntraits)
        >>> C0 = C0 @ C0.T
        >>>
        >>> C1 = random.randn(ntraits, ntraits)
        >>> C1 = C1 @ C1.T
        >>>
        >>> G = random.randn(n, 4)
        >>>
        >>> A0 = random.randn(ntraits, 1)
        >>> A1 = random.randn(ntraits, 2)
        >>> A01 = concatenate((A0, A1), axis=1)
        >>>
        >>> K = random.randn(n, n + 1)
        >>> K = normalise_covariance(K @ K.T)
        >>>
        >>> beta = vec(random.randn(ntraits, ncovariates))
        >>> alpha = vec(random.randn(A01.shape[1], G.shape[1]))
        >>>
        >>> mvn = st.multivariate_normal
        >>> m = kron(A, M) @ beta + kron(A01, G) @ alpha
        >>> Y = unvec(mvn(m, kron(C0, K) + kron(C1, eye(n))).rvs(), (n, -1))
        >>>
        >>> idx = [[0, 1], 2, [3]]
        >>> r = scan(G, Y, idx=idx, K=K, M=M, A=A, A0=A0, A1=A1, verbose=False)

    .. doctest::

        >>> from numpy import dot, exp, sqrt, ones
        >>> from numpy.random import RandomState
        >>> from pandas import DataFrame
        >>> import pandas as pd
        >>> from limix.qtl import scan
        >>>
        >>> random = RandomState(1)
        >>> pd.options.display.float_format = "{:9.6f}".format
        >>>
        >>> n = 30
        >>> p = 3
        >>> samples_index = range(n)
        >>>
        >>> M = DataFrame(dict(offset=ones(n), age=random.randint(10, 60, n)))
        >>> M.index = samples_index
        >>>
        >>> X = random.randn(n, 100)
        >>> K = dot(X, X.T)
        >>>
        >>> candidates = random.randn(n, p)
        >>> candidates = DataFrame(candidates, index=samples_index,
        ...                                    columns=['rs0', 'rs1', 'rs2'])
        >>>
        >>> y = random.poisson(exp(random.randn(n)))
        >>>
        >>> result = scan(candidates, y, 'poisson', K, M=M, verbose=False)
        >>>
        >>> result.stats  # doctest: +FLOAT_CMP +SKIP
               null lml    alt lml    pvalue  dof
        test
        0    -48.736563 -48.561855  0.554443    1
        1    -48.736563 -47.981093  0.218996    1
        2    -48.736563 -48.559868  0.552200    1
        >>> result.alt_effsizes  # doctest: +FLOAT_CMP +SKIP
           test candidate   effsize  effsize se
        0     0       rs0 -0.130867    0.221390
        1     1       rs1 -0.315079    0.256327
        2     2       rs2 -0.143869    0.242014
        >>> print(result)  # doctest: +FLOAT_CMP +SKIP
        Null model
        ----------
        <BLANKLINE>
          𝐳 ~ 𝓝(M𝜶, 0.79*K + 0.00*I)
          yᵢ ~ Poisson(λᵢ=g(zᵢ)), where g(x)=eˣ
          M = ['offset' 'age']
          𝜶 = [ 0.39528617 -0.00556789]
          Log marg. lik.: -48.736563230140376
          Number of models: 1
        <BLANKLINE>
        Alt model
        ---------
        <BLANKLINE>
          𝐳 ~ 𝓝(M𝜶 + Gᵢ, 0.79*K + 0.00*I)
          yᵢ ~ Poisson(λᵢ=g(zᵢ)), where g(x)=eˣ
          Min. p-value: 0.21899561824721903
          First perc. p-value: 0.22565970374303942
          Max. log marg. lik.: -47.981092939974765
          99th perc. log marg. lik.: -47.9926684371547
          Number of models: 3

        >>> from numpy import zeros
        >>>
        >>> nsamples = 50
        >>>
        >>> X = random.randn(nsamples, 2)
        >>> G = random.randn(nsamples, 100)
        >>> K = dot(G, G.T)
        >>> ntrials = random.randint(1, 100, nsamples)
        >>> z = dot(G, random.randn(100)) / sqrt(100)
        >>>
        >>> successes = zeros(len(ntrials), int)
        >>> for i, nt in enumerate(ntrials):
        ...     for _ in range(nt):
        ...         successes[i] += int(z[i] + 0.5 * random.randn() > 0)
        >>>
        >>> result = scan(X, successes, ("binomial", ntrials), K, verbose=False)
        >>> print(result)  # doctest: +FLOAT_CMP +SKIP
        Null model
        ----------
        <BLANKLINE>
          𝐳 ~ 𝓝(M𝜶, 1.74*K + 0.15*I)
          yᵢ ~ Binom(μᵢ=g(zᵢ), nᵢ), where g(x)=1/(1+e⁻ˣ)
          M = ['offset']
          𝜶 = [0.40956947]
          Log marg. lik.: -142.9436437096321
          Number of models: 1
        <BLANKLINE>
        Alt model
        ---------
        <BLANKLINE>
          𝐳 ~ 𝓝(M𝜶 + Gᵢ, 1.74*K + 0.15*I)
          yᵢ ~ Binom(μᵢ=g(zᵢ), nᵢ), where g(x)=1/(1+e⁻ˣ)
          Min. p-value: 0.23699422686919802
          First perc. p-value: 0.241827874774993
          Max. log marg. lik.: -142.24445140459548
          99th perc. log marg. lik.: -142.25080258276773
          Number of models: 2

    Notes
    -----
    It will raise a ``ValueError`` exception if non-finite values are passed. Please,
    refer to the :func:`limix.qc.mean_impute` function for missing value imputation.
    """
    from numpy_sugar.linalg import economic_qs

    lik = normalize_likelihood(lik)

    if A is None:
        if A0 is not None or A1 is not None:
            raise ValueError("You cannot define `A0` or `A1` without defining `A`.")

    with session_block("QTL analysis", disable=not verbose):

        with session_line("Normalising input... ", disable=not verbose):
            data = conform_dataset(Y, M, G=G, K=K)

        Y = data["y"]
        M = data["M"]
        G = data["G"]
        K = data["K"]

        assert_finite(Y, M, K)

        if K is not None:
            QS = economic_qs(K)
        else:
            QS = None

        if A is None:
            r = _single_trait_scan(idx, lik, Y, M, G, QS, verbose)
        else:
            r = _multi_trait_scan(idx, lik, Y, M, G, QS, A, A0, A1, verbose)

        r = r.create()
        if verbose:
            print(r)

        return r


def _single_trait_scan(idx, lik, Y, M, G, QS, verbose):
    from numpy import asarray

    if lik[0] == "normal":
        scanner, v0, v1 = _st_lmm(Y.values.ravel(), M.values, QS, verbose)
    else:
        scanner, v0, v1 = _st_glmm(Y.values.ravel(), lik, M.values, QS, verbose)
    pass

    r = STScanResultFactory(
        lik[0],
        Y.trait.item(),
        M.covariate,
        G.candidate,
        scanner.null_lml(),
        scanner.null_beta,
        scanner.null_beta_se,
        v0,
        v1,
    )

    if idx is None:
        r1 = scanner.fast_scan(G, verbose)
        for i in range(G.shape[1]):
            h2 = _normalise_scan_names({k: v[i] for k, v in r1.items()})
            r.add_test(i, h2)
    else:
        for i in idx:
            i = _2d_sel(i)
            h2 = _normalise_scan_names(scanner.scan(asarray(G[:, i], float)))
            r.add_test(i, h2)
    return r


def _multi_trait_scan(idx, lik, Y, M, G, QS, A, A0, A1, verbose):
    from xarray import concat
    from numpy import eye, asarray, empty

    ntraits = Y.shape[1]

    if A1 is None:
        A1 = eye(ntraits)

    if A0 is None:
        A0 = empty((ntraits, 0))

    A0 = _asarray(A0, "env0", ["sample", "env"])
    A1 = _asarray(A1, "env1", ["sample", "env"])
    A01 = concat([A0, A1], dim="env")

    if lik[0] == "normal":
        scanner, C0, C1 = _mt_lmm(Y, A, M, QS, verbose)
    else:
        msg = "Non-normal likelihood inference has not been implemented for"
        msg += " multiple traits yet."
        raise ValueError(msg)

    r = MTScanResultFactory(
        lik[0],
        Y.trait,
        M.covariate,
        G.candidate,
        A0.env,
        A1.env,
        scanner.null_lml(),
        scanner.null_beta,
        scanner.null_beta_se,
        C0,
        C1,
    )

    if idx is None:
        idx = range(G.shape[1])
    for i in idx:

        i = _2d_sel(i)
        g = asarray(G[:, i], float)

        if A0.shape[1] == 0:
            h1 = None
        else:
            h1 = _normalise_scan_names(scanner.scan(A0, g))

        h2 = _normalise_scan_names(scanner.scan(A01, g))
        r.add_test(i, h1, h2)

    return r


def _st_lmm(Y, M, QS, verbose):
    from numpy import nan
    from glimix_core.lmm import LMM

    lmm = LMM(Y, M, QS, restricted=False)
    lmm.fit(verbose=verbose)
    sys.stdout.flush()

    if QS is None:
        v0 = nan
    else:
        v0 = lmm.v0

    v1 = lmm.v1

    return lmm.get_fast_scanner(), v0, v1


def _st_glmm(y, lik, M, QS, verbose):
    from numpy import nan
    from glimix_core.glmm import GLMMExpFam, GLMMNormal

    glmm = GLMMExpFam(y, lik, M, QS)

    glmm.fit(verbose=verbose)

    if QS is None:
        v0 = nan
    else:
        v0 = glmm.v0

    v1 = glmm.v1
    sys.stdout.flush()

    eta = glmm.site.eta
    tau = glmm.site.tau

    gnormal = GLMMNormal(eta, tau, M, QS)
    gnormal.fit(verbose=verbose)

    return gnormal.get_fast_scanner(), v0, v1


def _mt_lmm(Y, A, M, QS, verbose):
    from glimix_core.lmm import Kron2Sum
    from numpy_sugar.linalg import ddot
    from numpy import sqrt, zeros

    if QS is None:
        KG = zeros((Y.shape[0], 1))
    else:
        KG = ddot(QS[0][0], sqrt(QS[1]))

    lmm = Kron2Sum(Y.values, A, M.values, KG, restricted=False)
    lmm.fit(verbose=verbose)
    sys.stdout.flush()

    C0 = lmm.C0
    C1 = lmm.C1

    return lmm.get_fast_scanner(), C0, C1


def _2d_sel(idx):
    from collections.abc import Iterable

    if not isinstance(idx, (slice, Iterable)):
        return [idx]

    return idx


def _normalise_scan_names(r):
    return {
        "lml": r["lml"],
        "covariate_effsizes": r["effsizes0"],
        "covariate_effsizes_se": r["effsizes0_se"],
        "candidate_effsizes": r["effsizes1"],
        "candidate_effsizes_se": r["effsizes1_se"],
        "scale": r["scale"],
    }



"""
_data module
============

Private subpackage for standardisation of data structures and names.
"""
from ._asarray import asarray
from ._assert import assert_filetype, assert_target
from ._conf import CONF
from ._conform import conform_dataset
from ._lik import normalize_likelihood

__all__ = [
    "CONF",
    "asarray",
    "assert_filetype",
    "assert_target",
    "conform_dataset",
    "normalize_likelihood",
]



from numpy import ones, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose

from limix.qc import remove_dependent_cols


def test_remove_dependent_cols():
    random = RandomState(0)

    X = random.randn(4, 5)
    X[:, 2] = 3 * X[:, 1]
    D = X[:, [0, 1, 3, 4]]
    assert_allclose(remove_dependent_cols(X), D)

    X = random.randn(4, 1)
    R = X.copy()
    assert_allclose(remove_dependent_cols(X), R)

    X = ones((3, 4))
    R = ones((3, 1))
    assert_allclose(remove_dependent_cols(X), R)

    X = zeros((3, 4))
    R = zeros((3, 0))
    assert_allclose(remove_dependent_cols(X), R)



__all__ = ["return_none_if_none"]


def return_none_if_none(f):
    """ Return ``None`` if first argument is ``None``. """
    from functools import wraps

    @wraps(f)
    def new_func(x, *args, **kwargs):
        if x is None:
            return None
        return f(x, *args, **kwargs)

    return new_func



import click

from .download import download
from .estimate_kinship import estimate_kinship
from .extract import extract
from .qtl import qtl
from .remove import remove
from .see import see


def _get_version():
    import pkg_resources
    import re
    from os.path import realpath, dirname, join

    if __name__ == "__main__":
        filepath = join(dirname(realpath(__file__)), "..", "__init__.py")
        with open(filepath, "r") as f:
            content = f.read()
    else:
        content = pkg_resources.resource_string(__name__.split(".")[0], "__init__.py")
        content = content.decode()

    c = re.compile(r"__version__ *= *('[^']+'|\"[^\"]+\")")
    m = c.search(content)
    if m is None:
        return "unknown"
    return m.groups()[0][1:-1]


@click.group(name="limix", context_settings=dict(help_option_names=["-h", "--help"]))
@click.pass_context
@click.version_option(_get_version())
def cli(ctx):
    pass


cli.add_command(see)
cli.add_command(estimate_kinship)
cli.add_command(download)
cli.add_command(extract)
cli.add_command(qtl)
cli.add_command(remove)



from numpy import dot, exp, sqrt, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose

from limix.her import estimate


def test_heritability_estimate_binomial():
    random = RandomState(0)
    nsamples = 50

    G = random.randn(nsamples, 100)
    K = dot(G, G.T)
    ntrials = random.randint(1, 100, nsamples)
    z = dot(G, random.randn(100)) / sqrt(100)

    successes = zeros(len(ntrials), int)
    for i, nt in enumerate(ntrials):
        for _ in range(nt):
            successes[i] += int(z[i] + 0.5 * random.randn() > 0)

    assert_allclose(
        estimate(successes, ("binomial", ntrials), K, verbose=False),
        0.9999992083233082,
        rtol=1e-3,
    )


def test_heritability_estimate_poisson():
    random = RandomState(0)

    G = random.randn(50, 100)
    K = dot(G, G.T)
    z = dot(G, random.randn(100)) / sqrt(100)
    y = random.poisson(exp(z))

    assert_allclose(
        estimate(y, "poisson", K, verbose=False), 0.991766763337491, rtol=1e-3
    )


def test_heritability_estimate_normal():
    random = RandomState(0)

    G = random.randn(50, 100)
    K = dot(G, G.T)
    z = dot(G, random.randn(100)) / sqrt(100)
    y = z + 0.2 * random.randn(50)

    h2 = estimate(y, "normal", K, verbose=False)
    assert_allclose(h2, 0.9751053293095242, rtol=1e-5)



from collections import Counter

from .._bits.deco import return_none_if_none
from .._bits.xarray import set_coord
from ._asarray import asarray as _asarray
from ._conf import CONF

set_coord = return_none_if_none(set_coord)
_asarray = return_none_if_none(_asarray)


def conform_dataset(y, M=None, G=None, K=None):
    r""" Convert data types to DataArray.

    This is a fundamental function for :mod:`limix` as it standardise outcome,
    covariates, genotype, and kinship arrays into :class:`xarray.DataArray` data type.
    Data arrays are :mod:`numpy`/:mod:`dask` arrays with indexed coordinates,
    therefore generalising data frames from :mod:`pandas`. It allows for lazy loading of
    data via dask arrays. It also supports arrays with different dimensionality and
    types, mixture of indexed and non-indexed arrays, and repeated sample labels.

    Examples
    --------

    .. doctest::

        >>> from __future__ import unicode_literals
        >>> import pytest
        >>> from numpy.random import RandomState
        >>> from pandas import DataFrame
        >>> from xarray import DataArray
        >>> from limix._data import conform_dataset
        >>>
        >>> random = RandomState(0)
        >>>
        >>> y = random.randn(4)
        >>> y = DataFrame(y, index=["sample0", "sample0", "sample1", "sample2"])
        >>>
        >>> G = random.randn(5, 6)
        >>>
        >>> data = conform_dataset(y, G=G)
        >>> print(data["y"])
        <xarray.DataArray 'trait' (sample: 4, trait: 1)>
        array([[1.764052],
               [0.400157],
               [0.978738],
               [2.240893]])
        Coordinates:
          * sample   (sample) object 'sample0' 'sample0' 'sample1' 'sample2'
          * trait    (trait) int64 0
        >>> print(data["G"])
        <xarray.DataArray 'genotype' (sample: 4, candidate: 6)>
        array([[ 1.867558, -0.977278,  0.950088, -0.151357, -0.103219,  0.410599],
               [ 0.144044,  1.454274,  0.761038,  0.121675,  0.443863,  0.333674],
               [ 1.494079, -0.205158,  0.313068, -0.854096, -2.55299 ,  0.653619],
               [ 0.864436, -0.742165,  2.269755, -1.454366,  0.045759, -0.187184]])
        Coordinates:
          * sample   (sample) object 'sample0' 'sample0' 'sample1' 'sample2'
        Dimensions without coordinates: candidate
        >>> K = random.randn(3, 3)
        >>> K = K.dot(K.T)
        >>> K = DataArray(K)
        >>> K.coords["dim_0"] = ["sample0", "sample1", "sample2"]
        >>> K.coords["dim_1"] = ["sample0", "sample1", "sample2"]
        >>>
        >>> data = conform_dataset(y, K=K)
        >>> print(data["y"])
        <xarray.DataArray 'trait' (sample: 4, trait: 1)>
        array([[1.764052],
               [0.400157],
               [0.978738],
               [2.240893]])
        Coordinates:
          * sample   (sample) object 'sample0' 'sample0' 'sample1' 'sample2'
          * trait    (trait) int64 0
        >>> print(data["K"])
        <xarray.DataArray 'covariance' (sample_0: 4, sample_1: 4)>
        array([[ 1.659103,  1.659103, -0.850801, -1.956422],
               [ 1.659103,  1.659103, -0.850801, -1.956422],
               [-0.850801, -0.850801,  1.687126, -0.194938],
               [-1.956422, -1.956422, -0.194938,  6.027272]])
        Coordinates:
          * sample_0  (sample_0) object 'sample0' 'sample0' 'sample1' 'sample2'
          * sample_1  (sample_1) object 'sample0' 'sample0' 'sample1' 'sample2'
        >>> with pytest.raises(ValueError):
        ...     conform_dataset(y, G=G, K=K)
    """
    y = _asarray(y, "trait", CONF["data_dims"]["trait"])
    M = _asarray(M, "covariate", CONF["data_dims"]["covariate"])
    G = _asarray(G, "genotype", CONF["data_dims"]["genotype"])
    K = _asarray(K, "covariance", CONF["data_dims"]["covariance"])

    data = {"y": y, "M": M, "G": G, "K": K}
    data = {k: v for k, v in data.items() if v is not None}

    sample_dims = [
        t
        for t in [
            ("y", "sample"),
            ("M", "sample"),
            ("G", "sample"),
            ("K", "sample_0"),
            ("K", "sample_1"),
        ]
        if t[0] in data
    ]

    data = _fix_samples(data, sample_dims)
    for n in data.keys():
        data[n].name = CONF["varname_to_target"][n]

    nsamples = len(data["y"].coords["sample"])
    same_size = all(data[n].coords[d].size == nsamples for n, d in sample_dims)

    data = _fix_covariates(data, same_size)

    # We accept non-unique samples when all sample sizes are equal.
    # In the other cases, we check for uniqueness.
    if not same_size:
        _check_uniqueness(data, sample_dims)
        _match_samples(data, sample_dims)

    return {k: data.get(k, None) for k in ["y", "M", "G", "K"]}


def _default_covariates(samples):
    from numpy import ones, asarray
    from xarray import DataArray

    M = ones((samples.size, 1))
    M = DataArray(
        M,
        # encoding={"dtype": "float64"},
        dims=["sample", "covariate"],
        coords={"sample": samples, "covariate": asarray(["offset"], dtype=object)},
    )

    return M


def _default_sample_coords(n):
    return ["sample{}".format(j) for j in range(n)]


def _fix_samples(data, sample_dims):
    from .._bits.xarray import take
    from numpy import array_equal

    samples_list = [
        data[n].coords[d].values for n, d in sample_dims if d in data[n].coords
    ]
    nmin_samples = min(data[n].coords[d].size for n, d in sample_dims)

    if not samples_list:
        data["y"] = take(data["y"], slice(0, nmin_samples), "sample")
        data["y"].coords["sample"] = _default_sample_coords(data["y"].sample.size)
        samples_list.append(data["y"].coords["sample"].values)

    samples = samples_list[0]

    if len(samples_list) < len(sample_dims):
        if not all([array_equal(s, samples) for s in samples_list]):
            raise ValueError(
                "Please, check the provided sample labels in your arrays."
                + " There are some inconsistences between them."
            )

        for n, d in sample_dims:
            data[n] = take(data[n], slice(0, nmin_samples), d)
            data[n].coords[d] = samples[:nmin_samples]

    samples_list = [
        data[n].coords[d].values for n, d in sample_dims if d in data[n].coords
    ]

    valid_samples = _infer_samples_index(samples_list)
    for n, d in sample_dims:
        data[n] = set_coord(data[n], d, valid_samples)

    return data


def _infer_samples_index(samples_list):
    r""" Infer a list of sample labels that is compatible to the provided data.

    It uses :class:`collections.Counter` to count the number of repeated sample
    labels, and to provide set (bag) intersection that handles repeated elements.
    """
    from numpy import array_equal, asarray

    samples = samples_list[0]
    if all(array_equal(s, samples) for s in samples_list):
        return Counter(samples)

    samples_sets = [Counter(s) for s in samples_list]

    set_intersection = samples_sets[0]
    for ss in samples_sets[1:]:
        set_intersection = set_intersection & ss

    membership_size = [
        asarray([ss[si] for ss in samples_sets], int) for si in set_intersection
    ]

    valid_samples = Counter()

    for k in set_intersection.keys():
        if sum(membership_size[0] > 1) <= 1:
            valid_samples[k] = set_intersection[k]

    return valid_samples


def _fix_covariates(data, samples_same_size):
    from pandas import unique

    y = data["y"]

    # We accept non-unique samples when all sample sizes are equal.
    if samples_same_size:
        if "M" not in data:
            data["M"] = _default_covariates(y.sample.values)
    else:
        if "M" not in data:
            data["M"] = _default_covariates(unique(y.sample.values))

    return data


def _check_uniqueness(data, dims):
    from numpy import unique

    msg = "Non-unique sample ids are not allowed in the {} array"
    msg += " if the sample ids are not equal nor in the same order."

    for n, d in dims:
        if n == "y":
            continue
        idx = data[n].coords[d].values
        if len(unique(idx)) < len(idx):
            raise ValueError(msg.format(data[n].name))


def _match_samples(data, dims):
    inc_msg = "The provided trait and {} arrays are sample-wise incompatible."

    for n, d in dims:
        if n == "y":
            continue
        try:
            data[n] = data[n].sel(**{d: data["y"].coords["sample"].values})
        except IndexError as e:
            raise ValueError(str(e) + "\n\n" + inc_msg.format(data[n].name))

    return data



from limix._bits import dask, numpy, pandas, xarray


def quantile_gaussianize(X, axis=1, inplace=False):
    r"""Normalize a sequence of values via rank and Normal c.d.f.

    It defaults to column-wise normalization.

    Parameters
    ----------
    X : array_like
        Array of values.
    axis : int, optional
        Axis value. Defaults to `1`.
    inplace : bool, optional
        Defaults to `False`.

    Returns
    -------
    array_like
        Gaussian-normalized values.

    Examples
    --------
    .. doctest::

        >>> from limix.qc import quantile_gaussianize
        >>> from numpy import array_str
        >>>
        >>> qg = quantile_gaussianize([-1, 0, 2])
        >>> print(qg) # doctest: +FLOAT_CMP
        [-0.67448975  0.          0.67448975]
    """
    from numpy import issubdtype, integer, asarray

    if hasattr(X, "dtype") and issubdtype(X.dtype, integer):
        raise ValueError("Integer type is not supported.")

    if isinstance(X, (tuple, list)):
        if inplace:
            raise ValueError("Can't use `inplace=True` for {}.".format(type(X)))
        X = asarray(X, float)

    if numpy.is_array(X):
        X = _qg_numpy(X, axis, inplace)
    elif pandas.is_series(X):
        X = _qg_pandas_series(X, axis, inplace)
    elif pandas.is_dataframe(X):
        X = _qg_pandas_dataframe(X, axis, inplace)
    elif dask.is_array(X):
        X = _qg_dask_array(X, axis, inplace)
    elif dask.is_series(X):
        raise NotImplementedError()
    elif dask.is_dataframe(X):
        X = _qg_dask_dataframe(X, axis, inplace)
    elif xarray.is_dataarray(X):
        X = _qg_xarray_dataarray(X, axis, inplace)
    else:
        raise NotImplementedError()

    return X


def _qg_numpy(X, axis, inplace):
    from scipy.stats import norm
    from numpy import isfinite
    from numpy.ma import masked_invalid
    from numpy_sugar import nanrankdata
    from numpy import apply_along_axis
    import warnings

    orig_shape = X.shape
    if X.ndim == 1:
        X = X.reshape(orig_shape + (1,))

    if not inplace:
        X = X.copy()

    D = X.swapaxes(1, axis)
    D = masked_invalid(D)
    D *= -1

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        D = nanrankdata(D)
        D = D / (isfinite(D).sum(axis=0) + 1)
        D = apply_along_axis(norm.isf, 0, D)

    D = D.swapaxes(1, axis)
    X[:] = D

    return X.reshape(orig_shape)


def _qg_pandas_series(X, axis, inplace):
    if not inplace:
        X = X.copy()

    a = X.to_numpy()
    _qg_numpy(a, axis, True)
    X[:] = a

    return X


def _qg_pandas_dataframe(x, axis, inplace):
    if not inplace:
        x = x.copy()

    a = x.to_numpy()
    _qg_numpy(a, axis, True)
    x[:] = a

    return x


def _qg_dask_array(x, axis, inplace):
    import dask.array as da
    from scipy.stats import norm
    from numpy_sugar import nanrankdata

    if inplace:
        raise NotImplementedError()

    x = x.swapaxes(1, axis)

    x = dask.array_shape_reveal(x)
    shape = da.compute(*x.shape)
    x = da.ma.masked_array(x)
    x *= -1
    x = da.apply_along_axis(_dask_apply, 0, x, nanrankdata, shape[0])
    x = x / (da.isfinite(x).sum(axis=0) + 1)
    x = da.apply_along_axis(_dask_apply, 0, x, norm.isf, shape[0])

    return x.swapaxes(1, axis)


def _qg_dask_dataframe(x, axis, inplace):
    if inplace:
        raise NotImplementedError()

    d = x.to_dask_array(lengths=True)
    orig_chunks = d.chunks
    d = _qg_dask_array(d, axis, False).rechunk(orig_chunks)
    return d.to_dask_dataframe(columns=x.columns, index=x.index)


def _qg_xarray_dataarray(X, axis, inplace):
    if not inplace:
        X = X.copy(deep=True)

    data = X.data

    if dask.is_array(data):
        data = _qg_dask_array(data, axis, inplace)
    else:
        data = _qg_numpy(data, axis, inplace)

    X.data = data
    return X


def _dask_apply(x, func1d, length):
    from numpy import resize
    import warnings

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        x = func1d(x)

    return resize(x, length)



import dask.array as da
from numpy import nan
from numpy.random import RandomState
from numpy.testing import assert_array_equal

from limix.qc import count_missingness


def test_count_missingness():
    random = RandomState(0)
    X = random.randn(3, 4)
    X[0, 0] = nan
    X[0, 2] = nan
    X[2, 2] = nan
    X[:, 3] = nan
    assert_array_equal(count_missingness(X), [1, 0, 2, 3])
    X = da.from_array(X, 2)
    assert_array_equal(count_missingness(X), [1, 0, 2, 3])



from collections import OrderedDict

from limix._display import session_line

from .._data import conform_dataset
from .._display import session_block


def st_sscan(G, y, E, M=None, tests=None, verbose=True):
    """Mixed-model with genetic effect heterogeneity.

    Parameters
    ----------
    pheno : (`N`, 1) ndarray
        phenotype data
    environments : (`N`, `E`) ndarray
        environments data.
    covs : (`N`, `D`) ndarray
        covariate design matrix.
        By default, ``covs`` is a (`N`, `1`) array of ones.
    tests : list
        Which tests are performed.
        Element list values are ``'inter'`` and ``'assoc'``.
        By default, only the interaction test is considered.
    rhos : list
        for the association test, a list of ``rho`` values must be specified.
        The choice of ``rho`` affects the statistical power of the test
        (for more information see the StructLMM paper).
        By default, ``rho=[0, 0.1**2, 0.2**2, 0.3**2, 0.4**2, 0.5**2, 0.5, 1.]``
    verbose : (bool, optional):
        if True, details such as runtime as displayed.
    """
    from struct_lmm import StructLMM
    from numpy import zeros, hstack, asarray
    from pandas import DataFrame

    rhos = [0.0, 0.1 ** 2, 0.2 ** 2, 0.3 ** 2, 0.4 ** 2, 0.5 ** 2, 0.5, 1.0]

    with session_block("struct-lmm analysis", disable=not verbose):

        with session_line("Normalising input... ", disable=not verbose):
            data = conform_dataset(y, M, G=G, K=None)

        y = data["y"]
        M = data["M"]
        G = data["G"]

        if tests is None:
            tests = ["inter"]

        if "inter" in tests:
            slmi = StructLMM(asarray(y, float), E, W=E, rho_list=[0])

        if "assoc" in tests:
            slmm = StructLMM(asarray(y, float), E, W=E, rho_list=rhos)
            slmm.fit_null(F=asarray(M, float), verbose=False)

        _pvi = zeros(G.shape[1])
        _pva = zeros(G.shape[1])
        for snp in range(G.shape[1]):
            x = asarray(G[:, [snp]], float)

            if "inter" in tests:
                # interaction test
                M1 = hstack((M, x))
                slmi.fit_null(F=M1, verbose=False)
                _pvi[snp] = slmi.score_2_dof(x)

            if "assoc" in tests:
                # association test
                _pva[snp] = slmm.score_2_dof(x)

    data = OrderedDict()
    data["pvi"] = _pvi
    data["pva"] = _pva
    return DataFrame(data)



def impute(data, spec):
    import limix
    from limix._data import CONF

    ncolons = sum(1 for c in spec if c == ":")
    if ncolons > 2:
        _syntax_error_msg("Impute", "<TARGET>:<DIM>:<METHOD>", spec)

    spec = spec.strip()
    spec = spec + ":" * (2 - ncolons)
    target, dim, method = [e.strip() for e in spec.split(":")]

    varname = CONF["target_to_varname"][target]
    x = data[varname]

    if dim == "":
        axis = -1
    else:
        axis = next(i for i, d in enumerate(x.dims) if d == dim)

    if method == "":
        method = "mean"

    if method == "mean":
        x = limix.qc.mean_impute(x, axis=axis)
    else:
        raise ValueError(f"Unrecognized impute method: {method}.")

    data[varname] = x

    return data


def normalize(data, spec):
    import limix
    from limix._data import CONF

    ncolons = sum(1 for c in spec if c == ":")
    if ncolons > 2:
        _syntax_error_msg("Normalize", "<TARGET>:<DIM>:<METHOD>", spec)

    spec = spec.strip()
    spec = spec + ":" * (2 - ncolons)
    target, dim, method = [e.strip() for e in spec.split(":")]

    varname = CONF["target_to_varname"][target]
    x = data[varname]

    if dim == "":
        axis = -1
    else:
        axis = next(i for i, d in enumerate(x.dims) if d == dim)

    if method == "":
        method = "gaussianize"

    if method == "gaussianize":
        x = limix.qc.quantile_gaussianize(x, axis=axis)
    elif method == "mean_std":
        x = limix.qc.mean_standardize(x, axis=axis)
    else:
        raise ValueError(f"Unrecognized normalization method: {method}.")

    data[varname] = x

    return data


def where(data, spec):
    from limix._bits.xarray import query
    from limix._data import CONF

    ncolons = sum(1 for c in spec if c == ":")
    if ncolons > 1:
        _syntax_error_msg("Where", "<TARGET>:<COND>", spec)

    spec = spec.strip()
    spec = spec + ":" * (1 - ncolons)
    target, cond = [e.strip() for e in spec.split(":")]

    varname = CONF["target_to_varname"][target]

    data[varname] = query(data[varname], cond)

    return data


def drop_missing(data, spec):
    from limix._data import CONF

    ncolons = sum(1 for c in spec if c == ":")
    if ncolons > 2:
        _syntax_error_msg("Drop-missing", "<TARGET>:<DIM>:<METHOD>", spec)

    spec = spec.strip()
    spec = spec + ":" * (2 - ncolons)
    target, dim, method = [e.strip() for e in spec.split(":")]

    if dim == "":
        if target == "trait":
            dim = "trait"
        elif target == "genotype":
            dim = "candidate"
        elif target == "covariate":
            dim = "covariate"

    if method == "":
        method = "any"

    varname = CONF["target_to_varname"][target]
    x = data[varname]

    if dim == "":
        dim = x.dims[-1]

    data[varname] = data[varname].dropna(dim, method)

    return data


def drop_maf(data, spec):
    raise NotImplementedError()


# def _process_filter_maf(maf, G):
#     from limix import compute_maf

#     mafs = compute_maf(G)
#     ok = mafs >= maf
#     return G.isel(candidate=ok)


def _syntax_error_msg(what, syntax, spec):
    msg = f"{what} syntax error. It should have been\n"
    msg += f"  {syntax}\n"
    msg += f"but we received `{spec}`."
    return msg
